/************************************************************************/
/*
dynamics for walking (two legs with knees, no feet but foot torques exist)
Lagrange approach for a floating body is combined with foot constraints.
LNE stands for Lagrange/Newton-Euler
*/
/***********************************************************************/
/* Notes:
1) accelerations are stored with old action
2) since set_state does not have an action, the accelerations are not
   meaningful after set_state().

/************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "useful/useful.h"
#include "nr2d/nrutil.h"
#include "dm/dm.h"
#include "dynamics.h"

/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
// Mathematica vesions of function calls

double Sin( double x )
{
  return sin( x );
}

double Cos( double x )
{
  return cos( x );
}

// CGA: handle special case of integer y more efficiently
double Power( double x, double y )
{
  if ( y == 2.0 )
    return x*x;

  return pow( x, y );
}

double Sqrt( double x )
{
  return sqrt( x );
}

double ArcSin( double x )
{
  return asin( x );
}

double ArcCos( double x )
{
  return acos( x );
}

// Note argument order switch
double ArcTan( double x, double y )
{
  return atan2( y, x );
}

double Csc( double x )
{
  return 1.0/sin( x );
}

double Cot( double x )
{
  return 1.0/tan( x );
}

/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
// used for debugging

void compare( char *s, double v1, double v2, double threshold )
{
  if ( fabs( v1 - v2 ) >= threshold )
    printf( "compare: %s: %g %g %g\n", s, v1, v2, v1-v2 );
}

/**********************************************************************/
// used for debugging

double compute_energy_lne( Dynamics *d )
{
  double energy = 0;
  double scm, sl, ms, Is, thcm, thl, mth, Ith, tocm, tol, mto, Ito;
  double xh, yh, ato, ahl, ahr, akl, akr;
  double xhd, yhd, atod, ahld, ahrd, akld, akrd;

  // use Mathematica variable names
  scm = d->shank_cm;
  sl = d->shank_length;
  ms = d->shank_mass;
  Is = d->shank_I;
  thcm = d->thigh_cm;
  thl = d->thigh_length;
  mth = d->thigh_mass;
  Ith = d->thigh_I;
  tocm = d->torso_cm;
  tol = d->torso_length;
  mto = d->torso_mass;
  Ito = d->torso_I;

  // use Mathematica variable names
  xh = d->full_state[X_XH];
  yh = d->full_state[X_YH];
  ato = d->full_state[X_ATO];
  ahl = d->full_state[X_AHL];
  ahr = d->full_state[X_AHR];
  akl = d->full_state[X_AKL];
  akr = d->full_state[X_AKR];

  xhd = d->full_state[X_XHD];
  yhd = d->full_state[X_YHD];
  atod = d->full_state[X_ATOD];
  ahld = d->full_state[X_AHLD];
  ahrd = d->full_state[X_AHRD];
  akld = d->full_state[X_AKLD];
  akrd = d->full_state[X_AKRD];

  // energies

  d->lne_slkew =
    (Power(ahld - akld + atod,2)*Is)/2;
  d->lne_slkex =
   (ms*Power(xhd + (ahld + atod)*thl*Cos(ahl + ato) - 
	     (ahld - akld + atod)*(scm - sl)*Cos(ahl - akl + ato),2))/2;
  d->lne_slkey =
    (ms*Power(yhd + (ahld + atod)*thl*Sin(ahl + ato) - 
	      (ahld - akld + atod)*(scm - sl)*Sin(ahl - akl + ato),2))/2;

  d->lne_slke = d->lne_slkew + d->lne_slkex + d->lne_slkey;

  d->lne_slpe =
    G*ms*(yh - thl*Cos(ahl + ato) + (scm - sl)*Cos(ahl - akl + ato));

  d->lne_thlkew =
    (Power(ahld + atod,2)*Ith)/2;

  d->lne_thlkex =
    (mth*Power(xhd - (ahld + atod)*(thcm - thl)*Cos(ahl + ato),2))/2;

  d->lne_thlkey =
    (mth*Power(yhd - (ahld + atod)*(thcm - thl)*Sin(ahl + ato),2))/2;

  d->lne_thlke = d->lne_thlkew + d->lne_thlkex + d->lne_thlkey;

  d->lne_thlpe =
    G*mth*(yh + (thcm - thl)*Cos(ahl + ato));

  d->lne_srkew =
    (Power(ahrd - akrd + atod,2)*Is)/2.;

  d->lne_srkex =
    (ms*Power(xhd + (ahrd + atod)*thl*Cos(ahr + ato) +
	      (ahrd - akrd + atod)*(-scm + sl)*Cos(ahr - akr + ato),2))/2.;

  d->lne_srkey =
    (ms*Power(yhd + (ahrd + atod)*thl*Sin(ahr + ato) +
	      (ahrd - akrd + atod)*(-scm + sl)*Sin(ahr - akr + ato),2))/2.;

  d->lne_srke = d->lne_srkew + d->lne_srkex + d->lne_srkey;

  d->lne_srpe =
    G*ms*(yh - thl*Cos(ahr + ato) + (scm - sl)*Cos(ahr - akr + ato));

  d->lne_thrkew =
    (Power(ahrd + atod,2)*Ith)/2.;

  d->lne_thrkex =
    (mth*Power(xhd + (ahrd + atod)*(-thcm + thl)*Cos(ahr + ato),2))/2.;

  d->lne_thrkey =
    (mth*Power(yhd + (ahrd + atod)*(-thcm + thl)*Sin(ahr + ato),2))/2.;

  d->lne_thrke = d->lne_thrkew + d->lne_thrkex + d->lne_thrkey;

  d->lne_thrpe =
    G*mth*(yh + (thcm - thl)*Cos(ahr + ato));

  d->lne_tokew =
    (Power(atod,2)*Ito)/2;

  d->lne_tokex =
    (mto*Power(xhd - atod*tocm*Cos(ato),2))/2;

  d->lne_tokey =
    (mto*Power(yhd - atod*tocm*Sin(ato),2))/2;

  d->lne_toke = d->lne_tokew + d->lne_tokex + d->lne_tokey;

  d->lne_tope =
    G*mto*(yh + tocm*Cos(ato));

  d->lne_kew = d->lne_slkew + d->lne_thlkew
    + d->lne_srkew + d->lne_thrkew + d->lne_tokew;
  d->lne_kex = d->lne_slkex + d->lne_thlkex
    + d->lne_srkex + d->lne_thrkex + d->lne_tokex;
  d->lne_key = d->lne_slkey + d->lne_thlkey
    + d->lne_srkey + d->lne_thrkey + d->lne_tokey;
  d->lne_ke = d->lne_kew + d->lne_kex + d->lne_key;
  d->lne_pe = d->lne_slpe + d->lne_thlpe
    + d->lne_srpe + d->lne_thrpe + d->lne_tope;

  energy = d->lne_ke + d->lne_pe;

  return energy;
}

/**********************************************************************/
// used for debugging

void check_energies_lne( Dynamics *d )
{

  compare( "Energy", d->lne_energy, d->energy, 1e-7 );
  compare( "slpe", d->lne_slpe, d->slpe, 1e-7 );
  compare( "srpe", d->lne_srpe, d->srpe, 1e-7 );
  compare( "thlpe", d->lne_thlpe, d->thlpe, 1e-7 );
  compare( "thrpe", d->lne_thrpe, d->thrpe, 1e-7 );
  compare( "tope", d->lne_tope, d->tope, 1e-7 );
  compare( "slke", d->lne_slke, d->slke, 1e-7 );
  compare( "thlke", d->lne_thlke, d->thlke, 1e-7 );
  compare( "srke", d->lne_srke, d->srke, 1e-7 );
  compare( "thrke", d->lne_thrke, d->thrke, 1e-7 );
  compare( "toke", d->lne_toke, d->toke, 1e-7 );
}

/**********************************************************************/
/**********************************************************************/

double angular_momentum( Dynamics *d )
{
  double ms, Is, mth, Ith, mto, Ito;

  // use Mathematica variable names
  ms = d->shank_mass;
  Is = d->shank_I;
  mth = d->thigh_mass;
  Ith = d->thigh_I;
  mto = d->torso_mass;
  Ito = d->torso_I;

  d->angular_momentum = 0;
  d->angular_momentum += ms*((d->xsl - d->am_point[0])*d->ysld -
			     (d->ysl - d->am_point[1])*d->xsld)
    + Is*d->wsl;
  d->angular_momentum += ms*((d->xsr - d->am_point[0])*d->ysrd -
			     (d->ysr - d->am_point[1])*d->xsrd)
    + Is*d->wsr;
  d->angular_momentum += mth*((d->xthl - d->am_point[0])*d->ythld -
			      (d->ythl - d->am_point[1])*d->xthld)
    + Ith*d->wthl;
  d->angular_momentum += mth*((d->xthr - d->am_point[0])*d->ythrd -
			      (d->ythr - d->am_point[1])*d->xthrd)
    + Ith*d->wthr;
  d->angular_momentum += mto*((d->xto - d->am_point[0])*d->ytod -
			      (d->yto - d->am_point[1])*d->xtod)
    + Ito*d->wto;

  return d->angular_momentum;
}

/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/*
  This routine provides a standard view of the system, independent of
  which dynamics is being used.
  It assumes that d->full_state is set.
  The variables set by this subroutine are useful for use in controllers.
*/

void forward_kinematics( Dynamics *d )
{
  int i;
  double scm, sl, ms, Is, thcm, thl, mth, Ith, tocm, tol, mto, Ito;
  double xh, yh, ato, ahl, ahr, akl, akr;
  double xhd, yhd, atod, ahld, ahrd, akld, akrd;

  // use Mathematica variable names
  scm = d->shank_cm;
  sl = d->shank_length;
  ms = d->shank_mass;
  Is = d->shank_I;
  thcm = d->thigh_cm;
  thl = d->thigh_length;
  mth = d->thigh_mass;
  Ith = d->thigh_I;
  tocm = d->torso_cm;
  tol = d->torso_length;
  mto = d->torso_mass;
  Ito = d->torso_I;

  // use Mathematica variable names
  xh = d->full_state[X_XH];
  yh = d->full_state[X_YH];
  ato = d->full_state[X_ATO];
  ahl = d->ahl = d->full_state[X_AHL];
  ahr = d->ahr = d->full_state[X_AHR];
  akl = d->akl = d->full_state[X_AKL];
  akr = d->akr = d->full_state[X_AKR];

  xhd = d->full_state[X_XHD];
  yhd = d->full_state[X_YHD];
  atod = d->full_state[X_ATOD];
  ahld = d->ahld = d->full_state[X_AHLD];
  ahrd = d->ahrd = d->full_state[X_AHRD];
  akld = d->akld = d->full_state[X_AKLD];
  akrd = d->akrd = d->full_state[X_AKRD];

  // angles with respect to vertical.
  d->ato = ato;
  d->athl = ato + ahl;
  d->athr = ato + ahr;
  d->asl = d->athl - akl;
  d->asr = d->athr - akr;

  d->aal = d->asl;
  d->aar = d->asr;

  d->wto = atod;
  d->wthl = d->wto + ahld;
  d->wthr = d->wto + ahrd;
  d->wsl = d->wthl - akld;
  d->wsr = d->wthr - akrd;

  d->aald = d->wsl;
  d->aard = d->wsr;

  // do all the sin/cos calls
  /*
  sasl = sin( d->asl );
  casl = cos( d->asl );
  sasr = sin( d->asr );
  casr = cos( d->asr );
  sathl = sin( d->athl );
  cathl = cos( d->athl );
  sathr = sin( d->athr );
  cathr = cos( d->athr );
  sato = sin( d->ato );
  cato = cos( d->ato );
  */

  // As much as possible the following is raw Mathematica output

  // positions
  d->xh = xh;
  d->yh = yh;

  d->xto = xh - tocm*Sin(ato);
  d->yto = yh + tocm*Cos(ato);

  d->xhead = xh - tol*Sin(ato);
  d->yhead = yh + tol*Cos(ato);

  d->xthl = xh + (-thcm + thl)*Sin(ahl + ato);
  d->ythl = yh - (-thcm + thl)*Cos(ahl + ato);

  d->xkl = xh + thl*Sin(ahl + ato);
  d->ykl = yh - thl*Cos(ahl + ato);

  d->xsl = d->xkl + (-scm + sl)*Sin(ahl - akl + ato);
  d->ysl = d->ykl - (-scm + sl)*Cos(ahl - akl + ato);

  d->xal = d->xkl + sl*Sin(ahl - akl + ato);
  d->yal = d->ykl - sl*Cos(ahl - akl + ato);

  d->xthr = xh + (-thcm + thl)*Sin(ahr + ato);
  d->ythr = yh - (-thcm + thl)*Cos(ahr + ato);

  d->xkr = xh + thl*Sin(ahr + ato);
  d->ykr = yh - thl*Cos(ahr + ato);

  d->xsr = d->xkr + (-scm + sl)*Sin(ahr - akr + ato);
  d->ysr = d->ykr -(-scm + sl)*Cos(ahr - akr + ato);

  d->xar = d->xkr + sl*Sin(ahr - akr + ato);
  d->yar = d->ykr - sl*Cos(ahr - akr + ato);

  // velocity
  d->xhd = xhd;
  d->yhd = yhd;

  d->xtod = xhd - atod*tocm*Cos(ato);
  d->ytod = yhd - atod*tocm*Sin(ato);

  d->xheadd = xhd - atod*tol*Cos(ato);
  d->yheadd = yhd - atod*tol*Sin(ato);

  d->xthld = xhd + (ahld + atod)*(-thcm + thl)*Cos(ahl + ato);
  d->ythld = yhd + (ahld + atod)*(-thcm + thl)*Sin(ahl + ato);

  d->xkld = xhd + (ahld + atod)*thl*Cos(ahl + ato);
  d->ykld = yhd + (ahld + atod)*thl*Sin(ahl + ato);

  d->xsld = d->xkld + (ahld - akld + atod)*(-scm + sl)*Cos(ahl - akl + ato);
  d->ysld = d->ykld + (ahld - akld + atod)*(-scm + sl)*Sin(ahl - akl + ato);

  d->xald = d->xkld + (ahld - akld + atod)*sl*Cos(ahl - akl + ato);
  d->yald = d->ykld + (ahld - akld + atod)*sl*Sin(ahl - akl + ato);

  d->xthrd = xhd + (ahrd + atod)*(-thcm + thl)*Cos(ahr + ato); 
  d->ythrd = yhd + (ahrd + atod)*(-thcm + thl)*Sin(ahr + ato);

  d->xkrd = xhd + (ahrd + atod)*thl*Cos(ahr + ato);
  d->ykrd = yhd + (ahrd + atod)*thl*Sin(ahr + ato);

  d->xsrd = d->xkrd + (ahrd - akrd + atod)*(-scm + sl)*Cos(ahr - akr + ato);
  d->ysrd = d->ykrd + (ahrd - akrd + atod)*(-scm + sl)*Sin(ahr - akr + ato);

  d->xard = d->xkrd + (ahrd - akrd + atod)*sl*Cos(ahr - akr + ato);
  d->yard = d->ykrd + (ahrd - akrd + atod)*sl*Sin(ahr - akr + ato);

  // generate foot force sensor readings
  switch ( d->dynamics_state )
    {
    case D_AIR:
      d->s_fx_l = 0.0;
      d->s_fy_l = 0.0;
      d->s_m_l = 0.0;
      d->s_fx_r = 0.0;
      d->s_fy_r = 0.0;
      d->s_m_r = 0.0;
      break;
    case D_SSL:
      d->s_fx_l = d->nr_b[SSL_B_FGSLX+1];
      d->s_fy_l = d->nr_b[SSL_B_FGSLY+1];
      d->s_m_l = d->action[U_AAL];
      d->s_fx_r = 0.0;
      d->s_fy_r = 0.0;
      d->s_m_r = 0.0;
      break;
    case D_SSR:
      d->s_fx_l = 0.0;
      d->s_fy_l = 0.0;
      d->s_m_l = 0.0;
      d->s_fx_r = d->nr_b[SSR_B_FGSRX+1];
      d->s_fy_r = d->nr_b[SSR_B_FGSRY+1];
      d->s_m_r = d->action[U_AAR];
      break;
    case D_DS:
      d->s_fx_l = d->nr_b[DS_B_FGSLX+1];
      d->s_fy_l = d->nr_b[DS_B_FGSLY+1];
      d->s_m_l = d->action[U_AAL];
      d->s_fx_r = d->nr_b[DS_B_FGSRX+1];
      d->s_fy_r = d->nr_b[DS_B_FGSRY+1];
      d->s_m_r = d->action[U_AAR];
      break;
    default:
      fprintf( stderr,
	       "Unrecognized dynamics state %d in forward_kinematics.\n",
	       d->dynamics_state );
      exit( -1 );
      break;
    }

  // center of mass

  d->xcom = (ms*d->xsl + mth*d->xthl + ms*d->xsr
	     + mth*d->xthr + mto*d->xto)/d->total_mass;
  d->ycom = (ms*d->ysl + mth*d->ythl + ms*d->ysr
	     + mth*d->ythr + mto*d->yto)/d->total_mass;
  // d->dx = d->xcom - d->xr;

  d->xcomd = (ms*d->xsld + mth*d->xthld
	      + ms*d->xsrd + mth*d->xthrd + mto*d->xtod)/d->total_mass;
  d->ycomd = (ms*d->ysld + mth*d->ythld
	      + ms*d->ysrd + mth*d->ythrd + mto*d->ytod)/d->total_mass;
  // d->dxd = d->xcomd - d->xrd;

  // energies

  d->slke = d->wsl*Is*d->wsl/2 + ms*d->xsld*d->xsld/2 + ms*d->ysld*d->ysld/2;
  d->slpe = ms*G*d->ysl;

  d->thlke = d->wthl*Ith*d->wthl/2 + mth*d->xthld*d->xthld/2 + mth*d->ythld*d->ythld/2;
  d->thlpe = mth*G*d->ythl;

  d->srke = d->wsr*Is*d->wsr/2 + ms*d->xsrd*d->xsrd/2 + ms*d->ysrd*d->ysrd/2;
  d->srpe = ms*G*d->ysr;

  d->thrke = d->wthr*Ith*d->wthr/2 + mth*d->xthrd*d->xthrd/2 + mth*d->ythrd*d->ythrd/2;
  d->thrpe = mth*G*d->ythr;

  d->toke = d->wto*Ito*d->wto/2 + mto*d->xtod*d->xtod/2 + mto*d->ytod*d->ytod/2;
  d->tope = mto*G*d->yto;

  d->ke = d->slke + d->thlke + d->srke + d->thrke + d->toke;
  d->pe = d->slpe + d->thlpe + d->srpe + d->thrpe + d->tope;

  d->energy = d->ke + d->pe;

  // only for debugging
  // d->lne_energy = compute_energy_lne( d );
}

/**********************************************************************/
// set state, given d->xh, d->yh, d->ato, d->xal, d->yal, d->xar, d->yar
// and their derivatives
// assumes dynamics state set in d

void set_state( Dynamics *d )
{
  int i;
  double ll2, lr2;
  double sl, thl;
  double akl, akr, offsetl, offsetr, asl, asr, athl, athr;
  double cakl, cakr, sakl, sakr;
  double ll2d, lr2d, akld, akrd, offsetld, offsetrd, asld, asrd;
  double ll2dd, lr2dd, akldd, akrdd, offsetldd, offsetrdd, asldd, asrdd;

  sl = d->shank_length;
  thl = d->thigh_length;

  // positions

  d->full_state[X_XH] = d->xh;
  d->full_state[X_YH] = d->yh;
  d->full_state[X_ATO] = d->ato;

  ll2 = (d->xh - d->xal)*(d->xh - d->xal) + (d->yh - d->yal)*(d->yh - d->yal);
  lr2 = (d->xh - d->xar)*(d->xh - d->xar) + (d->yh - d->yar)*(d->yh - d->yar);

  /*
  printf( "set_state: %g %g %g; %g %g %g %g; %g %g\n",
	  d->xh, d->yh, d->ato, d->xal, d->yal, d->xar, d->yar, ll2, lr2 );
  */

  if ( ll2 > (sl + thl)*(sl + thl) || lr2 > (sl + thl)*(sl + thl) )
    {
      fprintf( stderr, "Requested leg length too long in set_state.\n" );
      exit( -1 );
    }

  cakl = (ll2 - sl*sl - thl*thl)/(2*sl*thl);
  cakr = (lr2 - sl*sl - thl*thl)/(2*sl*thl);

  akl = acos( cakl );
  akr = acos( cakr );

  sakl = sin( akl );
  sakr = sin( akr );

  offsetl = atan2( thl*sakl, sl + thl*cakl );
  offsetr = atan2( thl*sakr, sl + thl*cakr );

  asl = atan2( -(d->xh - d->xal), d->yh - d->yal ) - offsetl;
  asr = atan2( -(d->xh - d->xar), d->yh - d->yar ) - offsetr;

  athl = asl + akl;
  athr = asr + akr;

  d->full_state[X_AKL] = akl;
  d->full_state[X_AKR] = akr;
  d->full_state[X_AHL] = athl - d->ato;
  d->full_state[X_AHR] = athr - d->ato;

  // velocities

  d->full_state[X_XHD] = d->xhd;
  d->full_state[X_YHD] = d->yhd;
  d->full_state[X_ATOD] = d->wto;

  ll2d = 2*(d->xh - d->xal)*(d->xhd - d->xald)
    + 2*(d->yh - d->yal)*(d->yhd - d->yald);
  lr2d = 2*(d->xh - d->xar)*(d->xhd - d->xard)
    + 2*(d->yh - d->yar)*(d->yhd - d->yard);

  akld = ll2d/(-2*sl*thl*sakl);
  akrd = lr2d/(-2*sl*thl*sakr);

  offsetld = (akld*thl*(thl + sl*cakl))/ll2;
  offsetrd = (akrd*thl*(thl + sl*cakr))/lr2;

  asld = -offsetld
    + (-((d->xald - d->xhd)*(d->yal - d->yh))
       + (d->xal - d->xh)*(d->yald - d->yhd))/ll2;

  asrd = -offsetrd
    + (-((d->xard - d->xhd)*(d->yar - d->yh))
       + (d->xar - d->xh)*(d->yard - d->yhd))/lr2;

  d->full_state[X_AKLD] = akld;
  d->full_state[X_AKRD] = akrd;
  d->full_state[X_AHLD] = asld + akld - d->wto;
  d->full_state[X_AHRD] = asrd + akrd - d->wto;

  // accelerations

  d->accelerations[X_XH] = d->xhdd;
  d->accelerations[X_YH] = d->yhdd;
  d->accelerations[X_ATO] = d->wtod;

  /*
  printf( "%g %g %g\n",
	  d->accelerations[X_XH],
	  d->accelerations[X_YH],
	  d->accelerations[X_ATO] );
  */

  ll2dd = 2*(Power(d->xald - d->xhd,2)
	     + Power(d->yald - d->yhd,2)
	     + (d->xal - d->xh)*(d->xaldd - d->xhdd)
	     + (d->yal - d->yh)*(d->yaldd - d->yhdd));
  lr2dd = 2*(Power(d->xard - d->xhd,2)
	     + Power(d->yard - d->yhd,2)
	     + (d->xar - d->xh)*(d->xardd - d->xhdd)
	     + (d->yar - d->yh)*(d->yardd - d->yhdd));

  akldd = -(Csc(akl)*(ll2dd - ll2d*Cot(akl)*akld))/(2.*sl*thl);
  akrdd = -(Csc(akr)*(lr2dd - lr2d*Cot(akr)*akrd))/(2.*sl*thl);

  offsetldd = (Csc(akl)*((Power(ll2d,2) - ll2*ll2dd)*(thl + sl*Cos(akl)) + 
			 ll2*ll2d*(sl + thl*Cos(akl))*Csc(akl)*akld))
    /(2.*Power(ll2,2)*sl);
  offsetrdd = (Csc(akr)*((Power(lr2d,2) - lr2*lr2dd)*(thl + sl*Cos(akr)) + 
			 lr2*lr2d*(sl + thl*Cos(akr))*Csc(akr)*akrd))
    /(2.*Power(lr2,2)*sl);

  asldd = -offsetldd
    + (ll2d*((d->xald - d->xhd)*(d->yal - d->yh)
	     - (d->xal - d->xh)*(d->yald - d->yhd))
       + ll2*(-((d->xaldd - d->xhdd)*(d->yal - d->yh))
	      + (d->xal - d->xh)*(d->yaldd - d->yhdd)))
    /Power(ll2,2);

  asrdd = -offsetrdd
    + (lr2d*((d->xard - d->xhd)*(d->yar - d->yh)
	     - (d->xar - d->xh)*(d->yard - d->yhd))
       + lr2*(-((d->xardd - d->xhdd)*(d->yar - d->yh))
	      + (d->xar - d->xh)*(d->yardd - d->yhdd)))
    /Power(lr2,2);

  /*
  printf( "%g %g %g %g %g %g %g %g\n", ll2dd, lr2dd, akldd, akrdd,
	  offsetldd, offsetrdd, asldd, asrdd);
  */

  d->accelerations[X_AKL] = akldd;
  d->accelerations[X_AKR] = akrdd;
  d->accelerations[X_AHL] = asldd + akldd - d->wtod;
  d->accelerations[X_AHR] = asrdd + akrdd - d->wtod; 

  /*
  printf( "%g %g %g %g\n", 
	  d->accelerations[X_AKL], d->accelerations[X_AKR],
	  d->accelerations[X_AHL], d->accelerations[X_AHR] );
  */

 /*
  printf( "xh, yh, ato: %g %g %g\n", d->full_state[X_XH],
	  d->full_state[X_YH], d->full_state[X_ATO] );
  printf( "xal, yal, xar, yar: %g %g %g %g\n",
	  d->xal, d->yal, d->xar, d->yar );
  printf( "ll2, lr2, ol, or: %g %g %g %g\n", ll2, lr2, offsetl, offsetr );
  printf( "akl, akr: %g %g\n", d->full_state[X_AKL],
	  d->full_state[X_AKR] );
  printf( "asl, asr: %g %g\n", asl, asr );
  printf( "athl, athr: %g %g\n", athl, athr );
  printf( "ahl, ahr: %g %g\n", d->full_state[X_AHL],
	  d->full_state[X_AHR] );
  */

  /*
  printf( "velocities: %g %g %g %g %g %g %g\n", 
	  d->full_state[X_XHD],
	  d->full_state[X_YHD],
	  d->full_state[X_ATOD],
	  d->full_state[X_AKLD],
	  d->full_state[X_AKRD],
	  d->full_state[X_AHLD],
	  d->full_state[X_AHRD] );
  */

  /*
  for ( i = 0; i < N_ACC; i++ )
    d->full_state[i + N_ACC] = 0;
  */

  forward_kinematics( d );

  /*
  printf( "xh, yh: %g %g\n", d->xh, d->yh );
  printf( "xkl, ykl, xkr, ykr: %g %g %g %g\n",
	  d->xkl, d->ykl, d->xkr, d->ykr );
  printf( "xal, yal, xar, yar: %g %g %g %g\n",
	  d->xal, d->yal, d->xar, d->yar );
  */
}

/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/

void form_matrices_air( Dynamics *d, double *action )
{
  int i, j, k;
  double scm, sl, ms, Is, thcm, thl, mth, Ith, tocm, tol, mto, Ito;
  double xh, yh, akl, ahl, akr, ahr, ato;
  double xhd, yhd, akld, ahld, akrd, ahrd, atod;
  int row;
  
  // use Mathematica variable names
  scm = d->shank_cm;
  sl = d->shank_length;
  ms = d->shank_mass;
  Is = d->shank_I;
  thcm = d->thigh_cm;
  thl = d->thigh_length;
  mth = d->thigh_mass;
  Ith = d->thigh_I;
  tocm = d->torso_cm;
  tol = d->torso_length;
  mto = d->torso_mass;
  Ito = d->torso_I;

  // use Mathematica variable names
  xh = d->full_state[X_XH];
  yh = d->full_state[X_YH];
  ato = d->full_state[X_ATO];
  ahl = d->full_state[X_AHL];
  ahr = d->full_state[X_AHR];
  akl = d->full_state[X_AKL];
  akr = d->full_state[X_AKR];

  xhd = d->full_state[X_XHD];
  yhd = d->full_state[X_YHD];
  atod = d->full_state[X_ATOD];
  ahld = d->full_state[X_AHLD];
  ahrd = d->full_state[X_AHRD];
  akld = d->full_state[X_AKLD];
  akrd = d->full_state[X_AKRD];

  // Numerical Recipes format: indices start at 1
  for ( i = 1; i <= N_A_B; i++ )
    {
      d->nr_b[i] = 0.0;
      for ( j = 1; j <= N_A_B; j++ )
	d->nr_a[i][j] = 0.0;
    }

  // Solving Ax = b

  row = 1; // Numerical Recipes format: indices start at 1

  // As much as possible the following is raw Mathematica output

  // BODY DYNAMICS
  // 1 XHDD
  d->nr_a[row][X_XH+1] = 2*(ms + mth) + mto;

  d->nr_a[row][X_YH+1] = 0;

  d->nr_a[row][X_ATO+1] = 
   -(mto*tocm*Cos(ato)) + (-(mth*thcm) + (ms + mth)*thl)*Cos(ahl + ato) + 
    (-(mth*thcm) + (ms + mth)*thl)*Cos(ahr + ato) - 
    ms*(scm - sl)*(Cos(ahl - akl + ato) + Cos(ahr - akr + ato));

  d->nr_a[row][X_AHL+1] =
   (-(mth*thcm) + (ms + mth)*thl)*Cos(ahl + ato) + 
    ms*(-scm + sl)*Cos(ahl - akl + ato);

  d->nr_a[row][X_AHR+1] =   
   (-(mth*thcm) + (ms + mth)*thl)*Cos(ahr + ato) + 
    ms*(-scm + sl)*Cos(ahr - akr + ato);

  d->nr_a[row][X_AKL+1] = 
    ms*(scm - sl)*Cos(ahl - akl + ato);

  d->nr_a[row][X_AKR+1] = 
    ms*(scm - sl)*Cos(ahr - akr + ato);
    
  d->nr_b[row] = 
    -(
   Power(atod,2)*mto*tocm*Sin(ato) + 
    Power(ahld + atod,2)*mth*(thcm - thl)*Sin(ahl + ato) - 
    Power(ahld + atod,2)*ms*thl*Sin(ahl + ato) + 
    Power(ahrd + atod,2)*mth*(thcm - thl)*Sin(ahr + ato) - 
    Power(ahrd + atod,2)*ms*thl*Sin(ahr + ato) + 
    Power(ahld - akld + atod,2)*ms*(scm - sl)*Sin(ahl - akl + ato) + 
    Power(ahrd - akrd + atod,2)*ms*(scm - sl)*Sin(ahr - akr + ato)
      );

  row++;

  // 2 YHDD
  d->nr_a[row][X_XH+1] = 0;

  d->nr_a[row][X_YH+1] = 2*(ms + mth) + mto;

  d->nr_a[row][X_ATO+1] =   
   -(mto*tocm*Sin(ato)) + (-(mth*thcm) + (ms + mth)*thl)*Sin(ahl + ato) + 
    (-(mth*thcm) + (ms + mth)*thl)*Sin(ahr + ato) - 
    ms*(scm - sl)*(Sin(ahl - akl + ato) + Sin(ahr - akr + ato));

  d->nr_a[row][X_AHL+1] =   
   (-(mth*thcm) + (ms + mth)*thl)*Sin(ahl + ato) + 
    ms*(-scm + sl)*Sin(ahl - akl + ato);

  d->nr_a[row][X_AHR+1] = 
   (-(mth*thcm) + (ms + mth)*thl)*Sin(ahr + ato) + 
    ms*(-scm + sl)*Sin(ahr - akr + ato);

  d->nr_a[row][X_AKL+1] = 
    ms*(scm - sl)*Sin(ahl - akl + ato);

  d->nr_a[row][X_AKR+1] = 
  ms*(scm - sl)*Sin(ahr - akr + ato);

  d->nr_b[row] = 
    -(
   G*(2*(ms + mth) + mto) - Power(atod,2)*mto*tocm*Cos(ato) + 
    Power(ahld + atod,2)*ms*thl*Cos(ahl + ato) + 
    Power(ahld + atod,2)*mth*(-thcm + thl)*Cos(ahl + ato) + 
    Power(ahrd + atod,2)*ms*thl*Cos(ahr + ato) + 
    Power(ahrd + atod,2)*mth*(-thcm + thl)*Cos(ahr + ato) - 
    Power(ahld - akld + atod,2)*ms*(scm - sl)*Cos(ahl - akl + ato) - 
    Power(ahrd - akrd + atod,2)*ms*(scm - sl)*Cos(ahr - akr + ato)
      );

  row++;

  // 3 ATODD
  d->nr_a[row][X_XH+1] = 
   -(mto*tocm*Cos(ato)) + (-(mth*thcm) + (ms + mth)*thl)*Cos(ahl + ato) + 
    (-(mth*thcm) + (ms + mth)*thl)*Cos(ahr + ato) - 
    ms*(scm - sl)*(Cos(ahl - akl + ato) + Cos(ahr - akr + ato));

  d->nr_a[row][X_YH+1] = 
   -(mto*tocm*Sin(ato)) + (-(mth*thcm) + (ms + mth)*thl)*Sin(ahl + ato) + 
    (-(mth*thcm) + (ms + mth)*thl)*Sin(ahr + ato) - 
    ms*(scm - sl)*(Sin(ahl - akl + ato) + Sin(ahr - akr + ato));

  d->nr_a[row][X_ATO+1] =
   2*Is + 2*Ith + Ito + 2*mth*Power(thcm - thl,2) + 
    2*ms*(Power(scm - sl,2) + Power(thl,2)) + mto*Power(tocm,2) - 
    2*ms*(scm - sl)*thl*(Cos(akl) + Cos(akr));

  d->nr_a[row][X_AHL+1] =   
    Is + Ith + mth*Power(thcm - thl,2) + 
    ms*(Power(scm - sl,2) + Power(thl,2)) + 2*ms*(-scm + sl)*thl*Cos(akl);

  d->nr_a[row][X_AHR+1] = 
   Is + Ith + mth*Power(thcm - thl,2) + 
    ms*(Power(scm - sl,2) + Power(thl,2)) + 2*ms*(-scm + sl)*thl*Cos(akr);

  d->nr_a[row][X_AKL+1] =
    -Is - ms*Power(scm - sl,2) + ms*(scm - sl)*thl*Cos(akl);

  d->nr_a[row][X_AKR+1] =
    -Is - ms*Power(scm - sl,2) + ms*(scm - sl)*thl*Cos(akr);

  d->nr_b[row] = 
    -(
   ms*(scm - sl)*thl*(akld*(2*ahld - akld + 2*atod)*Sin(akl) + 
       akrd*(2*ahrd - akrd + 2*atod)*Sin(akr)) + 
    G*(-(mto*tocm*Sin(ato)) + (-(mth*thcm) + (ms + mth)*thl)*Sin(ahl + ato) + 
       (-(mth*thcm) + (ms + mth)*thl)*Sin(ahr + ato) - 
       ms*(scm - sl)*(Sin(ahl - akl + ato) + Sin(ahr - akr + ato)))
      )
    ;

  row++;

  // 4 AHLDD
  d->nr_a[row][X_XH+1] = 
   (-(mth*thcm) + (ms + mth)*thl)*Cos(ahl + ato) + 
    ms*(-scm + sl)*Cos(ahl - akl + ato);

  d->nr_a[row][X_YH+1] =   
   (-(mth*thcm) + (ms + mth)*thl)*Sin(ahl + ato) + 
    ms*(-scm + sl)*Sin(ahl - akl + ato);

  d->nr_a[row][X_ATO+1] =   
   Is + Ith + mth*Power(thcm - thl,2) + 
    ms*(Power(scm - sl,2) + Power(thl,2)) + 2*ms*(-scm + sl)*thl*Cos(akl);

  d->nr_a[row][X_AHL+1] =   
   Is + Ith + mth*Power(thcm - thl,2) + 
    ms*(Power(scm - sl,2) + Power(thl,2)) + 2*ms*(-scm + sl)*thl*Cos(akl);

  d->nr_a[row][X_AHR+1] = 0;

  d->nr_a[row][X_AKL+1] =
    -Is - ms*Power(scm - sl,2) + ms*(scm - sl)*thl*Cos(akl);

  d->nr_a[row][X_AKR+1] = 0;

  d->nr_b[row] = 
    -(
   akld*(2*ahld - akld + 2*atod)*ms*(scm - sl)*thl*Sin(akl) + 
    G*(-(mth*thcm) + (ms + mth)*thl)*Sin(ahl + ato) + 
    G*ms*(-scm + sl)*Sin(ahl - akl + ato)
      )
    + action[U_AHL];

  row++;

  // 5 AHRDD
  d->nr_a[row][X_XH+1] = 
   (-(mth*thcm) + (ms + mth)*thl)*Cos(ahr + ato) + 
    ms*(-scm + sl)*Cos(ahr - akr + ato);

  d->nr_a[row][X_YH+1] =   
   (-(mth*thcm) + (ms + mth)*thl)*Sin(ahr + ato) + 
    ms*(-scm + sl)*Sin(ahr - akr + ato);

  d->nr_a[row][X_ATO+1] =   
   Is + Ith + mth*Power(thcm - thl,2) + 
    ms*(Power(scm - sl,2) + Power(thl,2)) + 2*ms*(-scm + sl)*thl*Cos(akr);

  d->nr_a[row][X_AHL+1] = 0;

  d->nr_a[row][X_AHR+1] =   
   Is + Ith + mth*Power(thcm - thl,2) + 
    ms*(Power(scm - sl,2) + Power(thl,2)) + 2*ms*(-scm + sl)*thl*Cos(akr);

  d->nr_a[row][X_AKL+1] = 0;

  d->nr_a[row][X_AKR+1] =
    -Is - ms*Power(scm - sl,2) + ms*(scm - sl)*thl*Cos(akr);

  d->nr_b[row] = 
    -(
   akrd*(2*ahrd - akrd + 2*atod)*ms*(scm - sl)*thl*Sin(akr) + 
    G*(-(mth*thcm) + (ms + mth)*thl)*Sin(ahr + ato) + 
    G*ms*(-scm + sl)*Sin(ahr - akr + ato)
      )
    + action[U_AHR];
  
  row++;

  // 6 AKLDD
  d->nr_a[row][X_XH+1] = 
    ms*(scm - sl)*Cos(ahl - akl + ato);

  d->nr_a[row][X_YH+1] = 
    ms*(scm - sl)*Sin(ahl - akl + ato);

  d->nr_a[row][X_ATO+1] =
    -Is - ms*Power(scm - sl,2) + ms*(scm - sl)*thl*Cos(akl);

  d->nr_a[row][X_AHL+1] =
    -Is - ms*Power(scm - sl,2) + ms*(scm - sl)*thl*Cos(akl);

  d->nr_a[row][X_AHR+1] = 0;

  d->nr_a[row][X_AKL+1] =
    Is + ms*Power(scm - sl,2);

  d->nr_a[row][X_AKR+1] = 0;

  d->nr_b[row] = 
    -(
      ms*(scm - sl)*(-(Power(ahld + atod,2)*thl*Sin(akl)) + 
		     G*Sin(ahl - akl + ato))
      )
    - action[U_AKL];

  row++;

  // 7 AKRDD
  d->nr_a[row][X_XH+1] = 
    ms*(scm - sl)*Cos(ahr - akr + ato);

  d->nr_a[row][X_YH+1] = 
    ms*(scm - sl)*Sin(ahr - akr + ato);

  d->nr_a[row][X_ATO+1] = 
    -Is - ms*Power(scm - sl,2) + ms*(scm - sl)*thl*Cos(akr);

  d->nr_a[row][X_AHL+1] = 0;

  d->nr_a[row][X_AHR+1] =
    -Is - ms*Power(scm - sl,2) + ms*(scm - sl)*thl*Cos(akr);

  d->nr_a[row][X_AKL+1] = 0;
  
  d->nr_a[row][X_AKR+1] = 
    Is + ms*Power(scm - sl,2);

  d->nr_b[row] = 
    -(
      ms*(scm - sl)*(-(Power(ahrd + atod,2)*thl*Sin(akr)) + 
		     G*Sin(ahr - akr + ato))
      )
    - action[U_AKR];

  row++;

  if ( row != N_AIR_B + 1 )
    {
      fprintf( stderr, "Row error: %d != %d\n", row, N_AIR_B + 1 );
      exit( -1 );
    }

  d->nr_n = N_AIR_B;

  // Debugging: Check for symmetry
  for ( j = 1; j <= N_AIR_B; j++ )
    {
      for ( k = j+1; k <= N_AIR_B; k++ )
	{
	  if ( fabs( d->nr_a[j][k] - d->nr_a[k][j] ) > 1e-12 )
	    {
	      printf( "%g: Asymmetric nr_a %d %d %g; %d %d %g; %g\n", d->time,
		      j, k, d->nr_a[j][k],
		      k, j, d->nr_a[k][j],
		      d->nr_a[j][k] - d->nr_a[k][j] );
	    }
	}
    }

  /*
  for ( i = 1; i <= N_AIR_B; i++ )
    {
      printf( "a%d: \n", i, d->nr_a[i][j] );
      for ( j = 1; j <= N_AIR_B; j++ )
	printf( " %g", d->nr_a[i][j] );
      printf( "\n" );
    }

  for ( i = 1; i <= N_AIR_B; i++ )
    printf( "bb2 %d: before %g\n", i, d->nr_b[i] );
  */
}

/**********************************************************************/
/**********************************************************************/

void form_matrices_ssl( Dynamics *d, double *action )
{
  int i, j, k;
  double scm, sl, ms, Is, thcm, thl, mth, Ith, tocm, tol, mto, Ito;
  double xh, yh, akl, ahl, akr, ahr, ato;
  double xhd, yhd, akld, ahld, akrd, ahrd, atod;
  int row;
  
  // use Mathematica variable names
  scm = d->shank_cm;
  sl = d->shank_length;
  ms = d->shank_mass;
  Is = d->shank_I;
  thcm = d->thigh_cm;
  thl = d->thigh_length;
  mth = d->thigh_mass;
  Ith = d->thigh_I;
  tocm = d->torso_cm;
  tol = d->torso_length;
  mto = d->torso_mass;
  Ito = d->torso_I;

  // use Mathematica variable names
  xh = d->full_state[X_XH];
  yh = d->full_state[X_YH];
  ato = d->full_state[X_ATO];
  ahl = d->full_state[X_AHL];
  ahr = d->full_state[X_AHR];
  akl = d->full_state[X_AKL];
  akr = d->full_state[X_AKR];

  xhd = d->full_state[X_XHD];
  yhd = d->full_state[X_YHD];
  atod = d->full_state[X_ATOD];
  ahld = d->full_state[X_AHLD];
  ahrd = d->full_state[X_AHRD];
  akld = d->full_state[X_AKLD];
  akrd = d->full_state[X_AKRD];

  form_matrices_air( d, action );

  row = 1; // Numerical Recipes format: indices start at 1

  // As much as possible the following is raw Mathematica output

  // BODY DYNAMICS
  // 1 XHDD
  d->nr_a[row][SSL_B_FGSLX+1] = -1.0;
  row++;

  // 2 YHDD
  d->nr_a[row][SSL_B_FGSLY+1] = -1.0;
  row++;

  // 3 ATODD
  d->nr_a[row][SSL_B_FGSLX+1] =
    -(thl*Cos(ahl + ato) + sl*Cos(ahl - akl + ato));
  d->nr_a[row][SSL_B_FGSLY+1] =
    -(thl*Sin(ahl + ato) + sl*Sin(ahl - akl + ato));
  d->nr_b[row] += -action[U_AAL];
  row++;

  // 4 AHLDD
  d->nr_a[row][SSL_B_FGSLX+1] =
    -(thl*Cos(ahl + ato) + sl*Cos(ahl - akl + ato));
  d->nr_a[row][SSL_B_FGSLY+1] =
    -(thl*Sin(ahl + ato) + sl*Sin(ahl - akl + ato));
  d->nr_b[row] += -action[U_AAL];
  row++;

  // 5 AHRDD
  row++;

  // 6 AKLDD
  d->nr_a[row][SSL_B_FGSLX+1] = -(-sl*Cos(ahl - akl + ato));
  d->nr_a[row][SSL_B_FGSLY+1] = -(-sl*Sin(ahl - akl + ato));
  d->nr_b[row] += action[U_AAL];
  row++;

  // 7 AKRDD
  row++;

  // BOARD/LEFT SHANK (ANKLE) CONSTRAINTS
  // 8 
  d->nr_a[row][X_XH+1] = -1.0;
  d->nr_a[row][X_ATO+1] = -(thl*Cos(ahl + ato) + sl*Cos(ahl - akl + ato));
  d->nr_a[row][X_AHL+1] = -(thl*Cos(ahl + ato) + sl*Cos(ahl - akl + ato));
  d->nr_a[row][X_AKL+1] = -(-sl*Cos(ahl - akl + ato));
  d->nr_b[row] = ( -Power(ahld + atod,2)*thl*Sin(ahl + ato)
			-  Power(ahld - akld + atod,2)*sl*Sin(ahl - akl + ato) );
  row++;

  // 9
  d->nr_a[row][X_YH+1] = -1.0;
  d->nr_a[row][X_ATO+1] = -(thl*Sin(ahl + ato) + sl*Sin(ahl - akl + ato));
  d->nr_a[row][X_AHL+1] = -(thl*Sin(ahl + ato) + sl*Sin(ahl - akl + ato));
  d->nr_a[row][X_AKL+1] = -(-sl*Sin(ahl - akl + ato));
  d->nr_b[row] = ( + Power(ahld + atod,2)*thl*Cos(ahl + ato)
			+ Power(ahld - akld + atod,2)*sl*Cos(ahl - akl + ato) );
  row++;

  if ( row != N_SSL_B + 1 )
    {
      fprintf( stderr, "Row error: %d != %d\n", row, N_SSL_B + 1 );
      exit( -1 );
    }

  d->nr_n = N_SSL_B;

  // Debugging: Check for symmetry
  for ( j = 1; j <= N_SSL_B; j++ )
    {
      for ( k = j+1; k <= N_SSL_B; k++ )
	{
	  if ( fabs( d->nr_a[j][k] - d->nr_a[k][j] ) > 1e-12 )
	    {
	      printf( "%g: Asymmetric nr_a %d %d %g; %d %d %g; %g\n", d->time,
		      j, k, d->nr_a[j][k],
		      k, j, d->nr_a[k][j],
		      d->nr_a[j][k] - d->nr_a[k][j] );
	    }
	}
    }

  /*
  for ( i = 1; i <= N_SSL_B; i++ )
    {
      printf( "a%d: \n", i, d->nr_a[i][j] );
      for ( j = 1; j <= N_SSL_B; j++ )
	printf( " %g", d->nr_a[i][j] );
      printf( "\n" );
    }

  for ( i = 1; i <= N_SSL_B; i++ )
    printf( "bb2 %d: before %g\n", i, d->nr_b[i] );
  */
}

/**********************************************************************/
/**********************************************************************/

void form_matrices_ssr( Dynamics *d, double *action )
{
  int i, j, k;
  double scm, sl, ms, Is, thcm, thl, mth, Ith, tocm, tol, mto, Ito;
  double xh, yh, akl, ahl, akr, ahr, ato;
  double xhd, yhd, akld, ahld, akrd, ahrd, atod;
  int row;
  
  // use Mathematica variable names
  scm = d->shank_cm;
  sl = d->shank_length;
  ms = d->shank_mass;
  Is = d->shank_I;
  thcm = d->thigh_cm;
  thl = d->thigh_length;
  mth = d->thigh_mass;
  Ith = d->thigh_I;
  tocm = d->torso_cm;
  tol = d->torso_length;
  mto = d->torso_mass;
  Ito = d->torso_I;

  // use Mathematica variable names
  xh = d->full_state[X_XH];
  yh = d->full_state[X_YH];
  ato = d->full_state[X_ATO];
  ahl = d->full_state[X_AHL];
  ahr = d->full_state[X_AHR];
  akl = d->full_state[X_AKL];
  akr = d->full_state[X_AKR];

  xhd = d->full_state[X_XHD];
  yhd = d->full_state[X_YHD];
  atod = d->full_state[X_ATOD];
  ahld = d->full_state[X_AHLD];
  ahrd = d->full_state[X_AHRD];
  akld = d->full_state[X_AKLD];
  akrd = d->full_state[X_AKRD];

  form_matrices_air( d, action );

  row = 1; // Numerical Recipes format: indices start at 1

  // As much as possible the following is raw Mathematica output

  // BODY DYNAMICS
  // 1 XHDD
  d->nr_a[row][SSR_B_FGSRX+1] = -1.0;
  row++;

  // 2 YHDD
  d->nr_a[row][SSR_B_FGSRY+1] = -1.0;
  row++;

  // 3 ATODD
  d->nr_a[row][SSR_B_FGSRX+1] =
    -(thl*Cos(ahr + ato) + sl*Cos(ahr - akr + ato));
  d->nr_a[row][SSR_B_FGSRY+1] =
    -(thl*Sin(ahr + ato) + sl*Sin(ahr - akr + ato));
  d->nr_b[row] += -action[U_AAR];
  row++;

  // 4 AHLDD
  row++;

  // 5 AHRDD
  d->nr_a[row][SSR_B_FGSRX+1] = -(thl*Cos(ahr + ato) + sl*Cos(ahr - akr + ato));
  d->nr_a[row][SSR_B_FGSRY+1] = -(thl*Sin(ahr + ato) + sl*Sin(ahr - akr + ato));
  d->nr_b[row] += -action[U_AAR];
  row++;

  // 6 AKLDD
  row++;

  // 7 AKRDD
  d->nr_a[row][SSR_B_FGSRX+1] = -(-sl*Cos(ahr - akr + ato));
  d->nr_a[row][SSR_B_FGSRY+1] = -(-sl*Sin(ahr - akr + ato));
  d->nr_b[row] += action[U_AAR];
  row++;

  // BOARD/RIGHT SHANK (ANKLE) CONSTRAINTS
  // 9 
  d->nr_a[row][X_XH+1] = -1.0;
  d->nr_a[row][X_ATO+1] = -(thl*Cos(ahr + ato) + sl*Cos(ahr - akr + ato));
  d->nr_a[row][X_AHR+1] = -(thl*Cos(ahr + ato) + sl*Cos(ahr - akr + ato));
  d->nr_a[row][X_AKR+1] = -(-sl*Cos(ahr - akr + ato));
  d->nr_b[row] = ( - Power(ahrd - akrd + atod,2)*sl*Sin(ahr - akr + ato)
			- Power(ahrd + atod,2)*thl*Sin(ahr + ato) );
  row++;

  // 9
  d->nr_a[row][X_YH+1] = -1.0;
  d->nr_a[row][X_ATO+1] = -(sl*Sin(ahr - akr + ato) + thl*Sin(ahr + ato));
  d->nr_a[row][X_AHR+1] = -(sl*Sin(ahr - akr + ato) + thl*Sin(ahr + ato));
  d->nr_a[row][X_AKR+1] = -(-sl*Sin(ahr - akr + ato));
  d->nr_b[row] = ( Power(ahrd + atod,2)*thl*Cos(ahr + ato)
			+ Power(ahrd - akrd + atod,2)*sl*Cos(ahr - akr + ato) );
  row++;

  if ( row != N_SSR_B + 1 )
    {
      fprintf( stderr, "Row error: %d != %d\n", row, N_SSR_B + 1 );
      exit( -1 );
    }

  d->nr_n = N_SSR_B;

  // Debugging: Check for symmetry
  for ( j = 1; j <= N_SSR_B; j++ )
    {
      for ( k = j+1; k <= N_SSR_B; k++ )
	{
	  if ( fabs( d->nr_a[j][k] - d->nr_a[k][j] ) > 1e-12 )
	    {
	      printf( "%g: Asymmetric nr_a %d %d %g; %d %d %g; %g\n", d->time,
		      j, k, d->nr_a[j][k],
		      k, j, d->nr_a[k][j],
		      d->nr_a[j][k] - d->nr_a[k][j] );
	    }
	}
    }

  /*
  for ( i = 1; i <= N_SSR_B; i++ )
    {
      printf( "a%d: \n", i, d->nr_a[i][j] );
      for ( j = 1; j <= N_SSR_B; j++ )
	printf( " %g", d->nr_a[i][j] );
      printf( "\n" );
    }

  for ( i = 1; i <= N_SSR_B; i++ )
    printf( "bb2 %d: before %g\n", i, d->nr_b[i] );
  */
}

/**********************************************************************/
/* could call ssl and ssr:
need to avoid duplicating calls to air
need to get L&R constraints in seperate lines
*/

void form_matrices_ds( Dynamics *d, double *action )
{
  int i, j, k;
  double scm, sl, ms, Is, thcm, thl, mth, Ith, tocm, tol, mto, Ito;
  double xh, yh, akl, ahl, akr, ahr, ato;
  double xhd, yhd, akld, ahld, akrd, ahrd, atod;
  int row;
  
  // use Mathematica variable names
  scm = d->shank_cm;
  sl = d->shank_length;
  ms = d->shank_mass;
  Is = d->shank_I;
  thcm = d->thigh_cm;
  thl = d->thigh_length;
  mth = d->thigh_mass;
  Ith = d->thigh_I;
  tocm = d->torso_cm;
  tol = d->torso_length;
  mto = d->torso_mass;
  Ito = d->torso_I;

  // use Mathematica variable names
  xh = d->full_state[X_XH];
  yh = d->full_state[X_YH];
  ato = d->full_state[X_ATO];
  ahl = d->full_state[X_AHL];
  ahr = d->full_state[X_AHR];
  akl = d->full_state[X_AKL];
  akr = d->full_state[X_AKR];

  xhd = d->full_state[X_XHD];
  yhd = d->full_state[X_YHD];
  atod = d->full_state[X_ATOD];
  ahld = d->full_state[X_AHLD];
  ahrd = d->full_state[X_AHRD];
  akld = d->full_state[X_AKLD];
  akrd = d->full_state[X_AKRD];

  form_matrices_air( d, action );

  row = 1; // Numerical Recipes format: indices start at 1

  // As much as possible the following is raw Mathematica output

  // BODY DYNAMICS
  // 1 XHDD
  d->nr_a[row][DS_B_FGSLX+1] = -1.0;
  d->nr_a[row][DS_B_FGSRX+1] = -1.0;

  row++;

  // 2 YHDD
  d->nr_a[row][DS_B_FGSLY+1] = -1.0;
  d->nr_a[row][DS_B_FGSRY+1] = -1.0;

  row++;

  // 3 ATODD
  d->nr_a[row][DS_B_FGSLX+1] = -(thl*Cos(ahl + ato) + sl*Cos(ahl - akl + ato));
  d->nr_a[row][DS_B_FGSLY+1] = -(thl*Sin(ahl + ato) + sl*Sin(ahl - akl + ato));
  d->nr_a[row][DS_B_FGSRX+1] = -(thl*Cos(ahr + ato) + sl*Cos(ahr - akr + ato));
  d->nr_a[row][DS_B_FGSRY+1] = -(thl*Sin(ahr + ato) + sl*Sin(ahr - akr + ato));

  d->nr_b[row] += -action[U_AAL] - action[U_AAR];

  row++;

  // 4 AHLDD
  d->nr_a[row][DS_B_FGSLX+1] = -(thl*Cos(ahl + ato) + sl*Cos(ahl - akl + ato));
  d->nr_a[row][DS_B_FGSLY+1] = -(thl*Sin(ahl + ato) + sl*Sin(ahl - akl + ato));

  d->nr_b[row] += -action[U_AAL];

  row++;

  // 5 AHRDD
  d->nr_a[row][DS_B_FGSRX+1] = -(thl*Cos(ahr + ato) + sl*Cos(ahr - akr + ato));
  d->nr_a[row][DS_B_FGSRY+1] = -(thl*Sin(ahr + ato) + sl*Sin(ahr - akr + ato));

  d->nr_b[row] += -action[U_AAR];
  
  row++;

  // 6 AKLDD
  d->nr_a[row][DS_B_FGSLX+1] = -(-sl*Cos(ahl - akl + ato));
  d->nr_a[row][DS_B_FGSLY+1] = -(-sl*Sin(ahl - akl + ato));

  d->nr_b[row] += action[U_AAL];

  row++;

  // 7 AKRDD
  d->nr_a[row][DS_B_FGSRX+1] = -(-sl*Cos(ahr - akr + ato));
  d->nr_a[row][DS_B_FGSRY+1] = -(-sl*Sin(ahr - akr + ato));

  d->nr_b[row] += action[U_AAR];

  row++;

  // BOARD/LEFT SHANK (ANKLE) CONSTRAINTS
  // 8 
  d->nr_a[row][X_XH+1] = -1.0;
  d->nr_a[row][X_ATO+1] = -(thl*Cos(ahl + ato) + sl*Cos(ahl - akl + ato));
  d->nr_a[row][X_AHL+1] = -(thl*Cos(ahl + ato) + sl*Cos(ahl - akl + ato));
  d->nr_a[row][X_AKL+1] = -(-sl*Cos(ahl - akl + ato));

  d->nr_b[row] = ( -Power(ahld + atod,2)*thl*Sin(ahl + ato)
			-  Power(ahld - akld + atod,2)*sl*Sin(ahl - akl + ato) );

  row++;

  // 9
  d->nr_a[row][X_YH+1] = -1.0;
  d->nr_a[row][X_ATO+1] = -(thl*Sin(ahl + ato) + sl*Sin(ahl - akl + ato));
  d->nr_a[row][X_AHL+1] = -(thl*Sin(ahl + ato) + sl*Sin(ahl - akl + ato));
  d->nr_a[row][X_AKL+1] = -(-sl*Sin(ahl - akl + ato));

  d->nr_b[row] = ( + Power(ahld + atod,2)*thl*Cos(ahl + ato)
			+ Power(ahld - akld + atod,2)*sl*Cos(ahl - akl + ato) );

  row++;

  // BOARD/RIGHT SHANK (ANKLE) CONSTRAINTS
  // 10 
  d->nr_a[row][X_XH+1] = -1.0;
  d->nr_a[row][X_ATO+1] = -(thl*Cos(ahr + ato) + sl*Cos(ahr - akr + ato));
  d->nr_a[row][X_AHR+1] = -(thl*Cos(ahr + ato) + sl*Cos(ahr - akr + ato));
  d->nr_a[row][X_AKR+1] = -(-sl*Cos(ahr - akr + ato));

  d->nr_b[row] = ( - Power(ahrd - akrd + atod,2)*sl*Sin(ahr - akr + ato)
			- Power(ahrd + atod,2)*thl*Sin(ahr + ato) );
  row++;

  // 11
  d->nr_a[row][X_YH+1] = -1.0;
  d->nr_a[row][X_ATO+1] = -(sl*Sin(ahr - akr + ato) + thl*Sin(ahr + ato));
  d->nr_a[row][X_AHR+1] = -(sl*Sin(ahr - akr + ato) + thl*Sin(ahr + ato));
  d->nr_a[row][X_AKR+1] = -(-sl*Sin(ahr - akr + ato));

  d->nr_b[row] = ( Power(ahrd + atod,2)*thl*Cos(ahr + ato)
			+ Power(ahrd - akrd + atod,2)*sl*Cos(ahr - akr + ato) );
  row++;

  if ( row != N_DS_B + 1 )
    {
      fprintf( stderr, "Row error: %d != %d\n", row, N_DS_B + 1 );
      exit( -1 );
    }

  d->nr_n = N_DS_B;

  // Debugging: Check for symmetry
  for ( j = 1; j <= N_DS_B; j++ )
    {
      for ( k = j+1; k <= N_DS_B; k++ )
	{
	  if ( fabs( d->nr_a[j][k] - d->nr_a[k][j] ) > 1e-12 )
	    {
	      printf( "%g: Asymmetric nr_a %d %d %g; %d %d %g; %g\n", d->time,
		      j, k, d->nr_a[j][k],
		      k, j, d->nr_a[k][j],
		      d->nr_a[j][k] - d->nr_a[k][j] );
	    }
	}
    }

  /*
  for ( i = 1; i <= N_DS_B; i++ )
    {
      printf( "a%d: \n", i, d->nr_a[i][j] );
      for ( j = 1; j <= N_DS_B; j++ )
	printf( " %g", d->nr_a[i][j] );
      printf( "\n" );
    }

  for ( i = 1; i <= N_DS_B; i++ )
    printf( "bb2 %d: before %g\n", i, d->nr_b[i] );
  */
}

/**********************************************************************/
/**********************************************************************/

void solve_Ax_b( Dynamics *d )
{
  int i;
  double nr_d;

  ludcmp( d->nr_a, d->nr_n, d->nr_indx, &nr_d );
  lubksb( d->nr_a, d->nr_n, d->nr_indx, d->nr_b );
}

/**********************************************************************/

void accelerations( Dynamics *d, double *action, double *xd )
{
  int i;

  switch ( d->dynamics_state )
    {
    case D_AIR:
      form_matrices_air( d, action );
      break;
    case D_SSL:
      form_matrices_ssl( d, action );
      break;
    case D_SSR:
      form_matrices_ssr( d, action );
      break;
    case D_DS:
      form_matrices_ds( d, action );
      break;
    default:
      fprintf( stderr, "Unrecognized dynamics state %d in accelerations.\n",
	       d->dynamics_state );
      exit( -1 );
      break;
    }

  solve_Ax_b( d );

  for ( i = 0; i < N_ACC; i++ )
    d->accelerations[i] = d->nr_b[i+1];

  if ( xd != NULL )
    {
      for ( i = 0; i < N_ACC; i++ )
	{
	  xd[i] = d->full_state[i + N_ACC];
	  xd[i + N_ACC] = d->nr_b[i + 1];
	}
    }
}

/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/

Dynamics *create_dynamics()
{
  int i;
  Dynamics *result;

  result = (Dynamics *) malloc( sizeof( Dynamics ) );
  if ( result == NULL )
    {
      fprintf( stderr, "Can't allocate dynamics.\n" );
      exit( -1 );
    }

  result->seed = 1; /* Set random number generator seed */

  result->time = 0;
  result->time_step = 0.0005;

  result->status = NOT_CRASHED;

  result->shank_cm = 0.25;
  result->shank_length = 0.5;
  result->shank_mass = 10.0;
  result->shank_I =
    result->shank_mass*result->shank_length*result->shank_length/12;
  result->thigh_cm = 0.25;
  result->thigh_length = 0.5;
  result->thigh_mass = 10.0;
  result->thigh_I =
    result->thigh_mass*result->thigh_length*result->thigh_length/12;
  result->torso_cm = 0.5;
  result->torso_length = 1.0;
  result->torso_mass = 50.0;
  result->torso_I =
    result->torso_mass*result->torso_length*result->torso_length/12;

  result->total_mass = 2*result->shank_mass + 2*result->thigh_mass
    + result->torso_mass;

  dv_zero( result->full_state, N_X );
  result->dynamics_state = D_UNKNOWN;
  dv_zero( result->desired_state, N_X );
  dv_zero( result->action, N_U );
  dv_zero( result->am_point, 3 );

  // Numerical recipes stuff; make sure big enough
  result->nr_a = matrix( 1, N_A_B, 1, N_A_B );

  result->lne_energy = 0.0;

  return result;
}

/**********************************************************************/

int policy( Dynamics *d, double *state, double *parameters, int index,
	    double *action )
{
  int iu, ix;
  int ind = 0;

  for ( iu = 0; iu < N_U; iu++ )
    {
      action[iu] = 0;
      for ( ix = 0; ix < N_X; ix++ )
	{
	  action[iu]
	    += parameters[ind++]*(state[ix] - d->desired_state[ix]);
	}
    }
}

/**********************************************************************/
// Approximate version using gradient descent
// assumes desired state is set in d.
// INPUT:
// qdd_d (length N_ACC) has the desired accelerations.
// forces_d (length N_DS_F) has the best guess as to the right forces
//          N_U actions followed by N_C constraint forces.
//          It can be zero.
// OUTPUT:
// forces_d (length N_DS_F) has the desired forces

#define N_ID_ITERATIONS 1000000
#define ID_DS_E2_THRESHOLD 1e-7
#define ID_DS_F2_THRESHOLD 1e-7

void inverse_dynamics_ds( Dynamics *d, double *qdd_d, double *forces_d )
{
  int i, j, k;
  double action[N_U];
  double forces[N_DS_F];
  double A_qdd[N_ACC]; // A*qdd_d
  double A_qdd_b[N_ACC]; // A*qdd_d - b
  double M[N_ACC][N_DS_F];
  double error[N_ACC];
  double weight[N_DS_F];
  double gradient[N_DS_F];
  double epsilon = 1e-1;
  double update[N_DS_F];
  double scm, sl, ms, Is, thcm, thl, mth, Ith, tocm, tol, mto, Ito;
  double xh, yh, akl, ahl, akr, ahr, ato;
  int row;
  double last_forces[N_DS_F];
  double error2;
  double df2;
  int count; 
  double vfe = 0; // vertical force equalizer

  // use zero actions to form A, b
  for ( i = 0; i < N_U; i++ )
    action[i] = 0;

  form_matrices_ds( d, action ); // form A, b

  // create initial guess as to forces.
  for ( i = 0; i < N_DS_F; i++ )
    last_forces[i] = forces[i] = forces_d[i];

  for ( i = 0; i < N_ACC; i++ )
    {
      A_qdd[i] = 0;
      for ( j = 0; j < N_ACC; j++ )
	A_qdd[i] += d->nr_a[i+1][j+1]*qdd_d[j];
      A_qdd_b[i] =  A_qdd[i] - d->nr_b[i+1];
    }

  /*
  printf( "A_qdd_b: " );
  for ( i = 0; i < N_ACC; i++ )
    printf( "%g ", A_qdd_b[i] );
  printf( "\n" );
  */

  // Now figure out M
  // use Mathematica variable names
  scm = d->shank_cm;
  sl = d->shank_length;
  ms = d->shank_mass;
  Is = d->shank_I;
  thcm = d->thigh_cm;
  thl = d->thigh_length;
  mth = d->thigh_mass;
  Ith = d->thigh_I;
  tocm = d->torso_cm;
  tol = d->torso_length;
  mto = d->torso_mass;
  Ito = d->torso_I;

  // use Mathematica variable names
  xh = d->full_state[X_XH];
  yh = d->full_state[X_YH];
  ato = d->full_state[X_ATO];
  ahl = d->full_state[X_AHL];
  ahr = d->full_state[X_AHR];
  akl = d->full_state[X_AKL];
  akr = d->full_state[X_AKR];

  // Numerical Recipes format: indices start at 1
  for ( i = 0; i < N_ACC; i++ )
    {
      for ( j = 0; j < N_DS_F; j++ )
	M[i][j] = 0.0;
    }

  row = 0; // Numerical Recipes format: indices start at 1

  // 1 XHDD
  M[row][DS_F_FGSLX] = -1.0;
  M[row][DS_F_FGSRX] = -1.0;

  row++;

  // 2 YHDD
  M[row][DS_F_FGSLY] = -1.0;
  M[row][DS_F_FGSRY] = -1.0;

  row++;

  // 3 ATODD
  M[row][DS_F_AAL] = 1.0;
  M[row][DS_F_AAR] = 1.0;

  M[row][DS_F_FGSLX] = -(thl*Cos(ahl + ato) + sl*Cos(ahl - akl + ato));
  M[row][DS_F_FGSLY] = -(thl*Sin(ahl + ato) + sl*Sin(ahl - akl + ato));
  M[row][DS_F_FGSRX] = -(thl*Cos(ahr + ato) + sl*Cos(ahr - akr + ato));
  M[row][DS_F_FGSRY] = -(thl*Sin(ahr + ato) + sl*Sin(ahr - akr + ato));

  row++;

  // 4 AHLDD
  M[row][DS_F_AAL] = 1.0;
  M[row][DS_F_AHL] = -1.0;

  M[row][DS_F_FGSLX] = -(thl*Cos(ahl + ato) + sl*Cos(ahl - akl + ato));
  M[row][DS_F_FGSLY] = -(thl*Sin(ahl + ato) + sl*Sin(ahl - akl + ato));

  row++;

  // 5 AHRDD
  M[row][DS_F_AAR] = 1.0;
  M[row][DS_F_AHR] = -1.0;

  M[row][DS_F_FGSRX] = -(thl*Cos(ahr + ato) + sl*Cos(ahr - akr + ato));
  M[row][DS_F_FGSRY] = -(thl*Sin(ahr + ato) + sl*Sin(ahr - akr + ato));

  row++;

  // 6 AKLDD
  M[row][DS_F_AAL] = -1.0;
  M[row][DS_F_AKL] = 1.0;

  M[row][DS_F_FGSLX] = -(-sl*Cos(ahl - akl + ato));
  M[row][DS_F_FGSLY] = -(-sl*Sin(ahl - akl + ato));

  row++;

  // 7 AKRDD
  M[row][DS_F_AAR] = -1.0;
  M[row][DS_F_AKR] = 1.0;

  M[row][DS_F_FGSRX] = -(-sl*Cos(ahr - akr + ato));
  M[row][DS_F_FGSRY] = -(-sl*Sin(ahr - akr + ato));

  row++;

  if ( row != N_ACC )
    {
      fprintf( stderr, "inverse_dynamics_ds: Row error: %d != %d\n",
	       row, N_ACC );
      exit( -1 );
    }

  /*
  printf( "M = [ " );
  for ( i = 0; i < row; i++ )
    {
      for ( j = 0; j < N_DS_F_ACTIONS_FORCES; j++ )
	printf( " %g", M[i][j] );
      printf( "\n" );
    }
  printf( "]\n" );
  */

  weight[DS_F_AHL] = 1.0;
  weight[DS_F_AKL] = 1.0;
  weight[DS_F_AAL] = 1e2;
  weight[DS_F_AHR] = 1.0;
  weight[DS_F_AKR] = 1.0;
  weight[DS_F_AAR] = 1e2;
  weight[DS_F_FGSLX] = 1e-7;
  weight[DS_F_FGSLY] = 1e-7;
  weight[DS_F_FGSRX] = 1e-7;
  weight[DS_F_FGSRY] = 1e-7;

  for ( count = 0; count < N_ID_ITERATIONS; count++ )
    {

      for ( i = 0; i < N_ACC; i++ )
	{
	  error[i] = 0;
	  for ( j = 0; j < (N_DS_F); j++ )
	    error[i] += M[i][j]*forces[j];
	  error[i] += A_qdd_b[i];
	}

      /*
      printf( "ID error: %d:", count );
      for ( j = 0; j < N_ACC; j++ )
	printf( " %g", error[j] );
      printf( "\n" );
      */

      error2 = 0.0;
      for ( j = 0; j < N_ACC; j++ )
	error2 += error[j]*error[j];
      if ( error2 < ID_DS_E2_THRESHOLD )
	break;

      for ( i = 0; i < (N_DS_F); i++ )
	{
	  gradient[i] = 0;
	  for ( j = 0; j < N_ACC; j++ )
	    gradient[i] += M[j][i]*error[j];
	  update[i] = -epsilon*(gradient[i] + (1e-3)*weight[i]*forces[i]);
	}
      vfe = -epsilon*(1e-5)*(forces[DS_F_FGSLY] - forces[DS_F_FGSRY]);
      update[DS_F_FGSLY] = vfe;
      update[DS_F_FGSRY] = -vfe;

      for ( i = 0; i < (N_DS_F); i++ )
	{
	  forces[i] += update[i];
	}

      df2 = 0.0;
      for ( j = 0; j < N_DS_F; j++ )
	df2 += (forces[j] - last_forces[j])*(forces[j] - last_forces[j]);
      if ( df2 < ID_DS_F2_THRESHOLD )
	break;
      for ( i = 0; i < N_DS_F; i++ )
	last_forces[i] = forces[i];
    }

  /*
  printf( "final forces: %d:", count );
  for ( j = 0; j < N_DS_F; j++ )
    printf( " %g", forces[j] );
  printf( "\n" );
  */

  for ( j = 0; j < N_DS_F; j++ )
    forces_d[j] = forces[j];
}

/**********************************************************************/
// Approximate version using gradient descent
// assumes desired state is set in d.
// INPUT:
// qdd_d (length N_ACC) has the desired accelerations.
// forces_d (length N_DS_F) has the best guess as to the right forces
//          N_U actions followed by N_C constraint forces.
//          It can be zero.
// OUTPUT:
// forces_d (length N_DS_F) has the desired forces (DS version)

void inverse_dynamics_ssl( Dynamics *d, double *qdd_d, double *forces_d )
{
  int i, j, k;
  double action[N_U];
  double forces[N_SSL_F];
  double A_qdd[N_ACC]; // A*qdd_d
  double A_qdd_b[N_ACC]; // A*qdd_d - b
  double M[N_ACC][N_SSL_F];
  double error[N_ACC];
  double weight[N_SSL_F];
  double gradient[N_SSL_F];
  double epsilon = 1e-1;
  double update[N_SSL_F];
  double scm, sl, ms, Is, thcm, thl, mth, Ith, tocm, tol, mto, Ito;
  double xh, yh, akl, ahl, akr, ahr, ato;
  int row;
  double last_forces[N_SSL_F];
  double error2;
  double df2;
  int count; 

  // use zero actions to form A, b
  for ( i = 0; i < N_U; i++ )
    action[i] = 0;

  form_matrices_ssl( d, action ); // form A, b

  // create initial guess as to forces.
  last_forces[SSL_F_AHL] = forces[SSL_F_AHL] = forces_d[DS_F_AHL];
  last_forces[SSL_F_AKL] = forces[SSL_F_AKL] = forces_d[DS_F_AKL];
  last_forces[SSL_F_AAL] = forces[SSL_F_AAL] = forces_d[DS_F_AAL];
  last_forces[SSL_F_AHR] = forces[SSL_F_AHR] = forces_d[DS_F_AHR];
  last_forces[SSL_F_AKR] = forces[SSL_F_AKR] = forces_d[DS_F_AKR];
  last_forces[SSL_F_FGSLX] = forces[SSL_F_FGSLX] = forces_d[DS_F_FGSLX];
  last_forces[SSL_F_FGSLY] = forces[SSL_F_FGSLY] = forces_d[DS_F_FGSLY];

  for ( i = 0; i < N_ACC; i++ )
    {
      A_qdd[i] = 0;
      for ( j = 0; j < N_ACC; j++ )
	A_qdd[i] += d->nr_a[i+1][j+1]*qdd_d[j];
      A_qdd_b[i] =  A_qdd[i] - d->nr_b[i+1];
    }

  /*
  printf( "A_qdd_b: " );
  for ( i = 0; i < N_ACC; i++ )
    printf( "%g ", A_qdd_b[i] );
  printf( "\n" );
  */

  // Now figure out M
  // use Mathematica variable names
  scm = d->shank_cm;
  sl = d->shank_length;
  ms = d->shank_mass;
  Is = d->shank_I;
  thcm = d->thigh_cm;
  thl = d->thigh_length;
  mth = d->thigh_mass;
  Ith = d->thigh_I;
  tocm = d->torso_cm;
  tol = d->torso_length;
  mto = d->torso_mass;
  Ito = d->torso_I;

  // use Mathematica variable names
  xh = d->full_state[X_XH];
  yh = d->full_state[X_YH];
  ato = d->full_state[X_ATO];
  ahl = d->full_state[X_AHL];
  ahr = d->full_state[X_AHR];
  akl = d->full_state[X_AKL];
  akr = d->full_state[X_AKR];

  // Numerical Recipes format: indices start at 1
  for ( i = 0; i < N_ACC; i++ )
    {
      for ( j = 0; j < N_SSL_F; j++ )
	M[i][j] = 0.0;
    }

  row = 0; // Numerical Recipes format: indices start at 1

  // 1 XHDD
  M[row][SSL_F_FGSLX] = -1.0;
  row++;

  // 2 YHDD
  M[row][SSL_F_FGSLY] = -1.0;
  row++;

  // 3 ATODD
  M[row][SSL_F_AAL] = 1.0;

  M[row][SSL_F_FGSLX] = -(thl*Cos(ahl + ato) + sl*Cos(ahl - akl + ato));
  M[row][SSL_F_FGSLY] = -(thl*Sin(ahl + ato) + sl*Sin(ahl - akl + ato));

  row++;

  // 4 AHLDD
  M[row][SSL_F_AAL] = 1.0;
  M[row][SSL_F_AHL] = -1.0;

  M[row][SSL_F_FGSLX] = -(thl*Cos(ahl + ato) + sl*Cos(ahl - akl + ato));
  M[row][SSL_F_FGSLY] = -(thl*Sin(ahl + ato) + sl*Sin(ahl - akl + ato));

  row++;

  // 5 AHRDD
  M[row][DS_F_AHR] = -1.0;
  row++;

  // 6 AKLDD
  M[row][SSL_F_AAL] = -1.0;
  M[row][SSL_F_AKL] = 1.0;

  M[row][SSL_F_FGSLX] = -(-sl*Cos(ahl - akl + ato));
  M[row][SSL_F_FGSLY] = -(-sl*Sin(ahl - akl + ato));

  row++;

  // 7 AKRDD
  M[row][DS_F_AKR] = 1.0;
  row++;

  if ( row != N_ACC )
    {
      fprintf( stderr, "inverse_dynamics_ssl: Row error: %d != %d\n",
	       row, N_ACC );
      exit( -1 );
    }

  /*
  printf( "M = [ " );
  for ( i = 0; i < row; i++ )
    {
      for ( j = 0; j < N_SSL_F_ACTIONS_FORCES; j++ )
	printf( " %g", M[i][j] );
      printf( "\n" );
    }
  printf( "]\n" );
  */

  weight[SSL_F_AHL] = 1e-7;
  weight[SSL_F_AKL] = 1e-7;
  weight[SSL_F_AAL] = 1e-7;
  weight[SSL_F_AHR] = 1e-7;
  weight[SSL_F_AKR] = 1e-7;
  weight[SSL_F_FGSLX] = 1e-7;
  weight[SSL_F_FGSLY] = 1e-7;

  for ( count = 0; count < N_ID_ITERATIONS; count++ )
    {

      for ( i = 0; i < N_ACC; i++ )
	{
	  error[i] = 0;
	  for ( j = 0; j < (N_SSL_F); j++ )
	    error[i] += M[i][j]*forces[j];
	  error[i] += A_qdd_b[i];
	}

      /*
      printf( "ID error: %d:", count );
      for ( j = 0; j < N_ACC; j++ )
	printf( " %g", error[j] );
      printf( "\n" );
      */

      error2 = 0.0;
      for ( j = 0; j < N_ACC; j++ )
	error2 += error[j]*error[j];
      if ( error2 < ID_DS_E2_THRESHOLD )
	break;

      for ( i = 0; i < (N_SSL_F); i++ )
	{
	  gradient[i] = 0;
	  for ( j = 0; j < N_ACC; j++ )
	    gradient[i] += M[j][i]*error[j];
	  update[i] = -epsilon*(gradient[i] + (1e-3)*weight[i]*forces[i]);
	}

      for ( i = 0; i < (N_SSL_F); i++ )
	{
	  forces[i] += update[i];
	}

      df2 = 0.0;
      for ( j = 0; j < N_SSL_F; j++ )
	df2 += (forces[j] - last_forces[j])*(forces[j] - last_forces[j]);
      if ( df2 < ID_DS_F2_THRESHOLD )
	break;
      for ( i = 0; i < N_SSL_F; i++ )
	last_forces[i] = forces[i];
    }

  /*
  printf( "final forces: %d:", count );
  for ( j = 0; j < N_SSL_F; j++ )
    printf( " %g", forces[j] );
  printf( "\n" );
  */

  forces_d[DS_F_AHL] = forces[SSL_F_AHL];
  forces_d[DS_F_AKL] = forces[SSL_F_AKL];
  forces_d[DS_F_AAL] = forces[SSL_F_AAL];
  forces_d[DS_F_AHR] = forces[SSL_F_AHR];
  forces_d[DS_F_AKR] = forces[SSL_F_AKR];
  forces_d[DS_F_AAR] = 0.0;
  forces_d[DS_F_FGSLX] = forces[SSL_F_FGSLX];
  forces_d[DS_F_FGSLY] = forces[SSL_F_FGSLY];
  forces_d[DS_F_FGSRX] = 0.0;
  forces_d[DS_F_FGSRY] = 0.0;
}

/**********************************************************************/
// Approximate version using gradient descent
// assumes desired state is set in d.
// INPUT:
// qdd_d (length N_ACC) has the desired accelerations.
// forces_d (length N_DS_F) has the best guess as to the right forces
//          N_U actions followed by N_C constraint forces.
//          It can be zero.
// OUTPUT:
// forces_d (length N_DS_F) has the desired forces (DS version)

void inverse_dynamics_ssr( Dynamics *d, double *qdd_d, double *forces_d )
{
  int i, j, k;
  double action[N_U];
  double forces[N_SSR_F];
  double A_qdd[N_ACC]; // A*qdd_d
  double A_qdd_b[N_ACC]; // A*qdd_d - b
  double M[N_ACC][N_SSR_F];
  double error[N_ACC];
  double weight[N_SSR_F];
  double gradient[N_SSR_F];
  double epsilon = 1e-1;
  double update[N_SSR_F];
  double scm, sl, ms, Is, thcm, thl, mth, Ith, tocm, tol, mto, Ito;
  double xh, yh, akl, ahl, akr, ahr, ato;
  int row;
  double last_forces[N_SSR_F];
  double error2;
  double df2;
  int count; 

  // use zero actions to form A, b
  for ( i = 0; i < N_U; i++ )
    action[i] = 0;

  form_matrices_ssr( d, action ); // form A, b

  // create initial guess as to forces.
  last_forces[SSR_F_AHL] = forces[SSR_F_AHL] = forces_d[DS_F_AHL];
  last_forces[SSR_F_AKL] = forces[SSR_F_AKL] = forces_d[DS_F_AKL];
  last_forces[SSR_F_AHR] = forces[SSR_F_AHR] = forces_d[DS_F_AHR];
  last_forces[SSR_F_AKR] = forces[SSR_F_AKR] = forces_d[DS_F_AKR];
  last_forces[SSR_F_AAR] = forces[SSR_F_AAR] = forces_d[DS_F_AAR];
  last_forces[SSR_F_FGSRX] = forces[SSR_F_FGSRX] = forces_d[DS_F_FGSRX];
  last_forces[SSR_F_FGSRY] = forces[SSR_F_FGSRY] = forces_d[DS_F_FGSRY];

  for ( i = 0; i < N_ACC; i++ )
    {
      A_qdd[i] = 0;
      for ( j = 0; j < N_ACC; j++ )
	A_qdd[i] += d->nr_a[i+1][j+1]*qdd_d[j];
      A_qdd_b[i] =  A_qdd[i] - d->nr_b[i+1];
    }

  /*
  printf( "A_qdd_b: " );
  for ( i = 0; i < N_ACC; i++ )
    printf( "%g ", A_qdd_b[i] );
  printf( "\n" );
  */

  // Now figure out M
  // use Mathematica variable names
  scm = d->shank_cm;
  sl = d->shank_length;
  ms = d->shank_mass;
  Is = d->shank_I;
  thcm = d->thigh_cm;
  thl = d->thigh_length;
  mth = d->thigh_mass;
  Ith = d->thigh_I;
  tocm = d->torso_cm;
  tol = d->torso_length;
  mto = d->torso_mass;
  Ito = d->torso_I;

  // use Mathematica variable names
  xh = d->full_state[X_XH];
  yh = d->full_state[X_YH];
  ato = d->full_state[X_ATO];
  ahl = d->full_state[X_AHL];
  ahr = d->full_state[X_AHR];
  akl = d->full_state[X_AKL];
  akr = d->full_state[X_AKR];

  // Numerical Recipes format: indices start at 1
  for ( i = 0; i < N_ACC; i++ )
    {
      for ( j = 0; j < N_SSR_F; j++ )
	M[i][j] = 0.0;
    }

  row = 0; // Numerical Recipes format: indices start at 1

  // 1 XHDD
  M[row][SSR_F_FGSRX] = -1.0;
  row++;

  // 2 YHDD
  M[row][SSR_F_FGSRY] = -1.0;
  row++;

  // 3 ATODD
  M[row][SSR_F_AAR] = 1.0;

  M[row][SSR_F_FGSRX] = -(thl*Cos(ahr + ato) + sl*Cos(ahr - akr + ato));
  M[row][SSR_F_FGSRY] = -(thl*Sin(ahr + ato) + sl*Sin(ahr - akr + ato));
  row++;

  // 4 AHLDD
  M[row][SSR_F_AHL] = -1.0;
  row++;

  // 5 AHRDD
  M[row][SSR_F_AAR] = 1.0;
  M[row][SSR_F_AHR] = -1.0;

  M[row][SSR_F_FGSRX] = -(thl*Cos(ahr + ato) + sl*Cos(ahr - akr + ato));
  M[row][SSR_F_FGSRY] = -(thl*Sin(ahr + ato) + sl*Sin(ahr - akr + ato));

  row++;

  // 6 AKLDD
  M[row][SSR_F_AKL] = 1.0;
  row++;

  // 7 AKRDD
  M[row][SSR_F_AAR] = -1.0;
  M[row][SSR_F_AKR] = 1.0;

  M[row][SSR_F_FGSRX] = -(-sl*Cos(ahr - akr + ato));
  M[row][SSR_F_FGSRY] = -(-sl*Sin(ahr - akr + ato));

  row++;

  if ( row != N_ACC )
    {
      fprintf( stderr, "inverse_dynamics_ssr: Row error: %d != %d\n",
	       row, N_ACC );
      exit( -1 );
    }

  /*
  printf( "M = [ " );
  for ( i = 0; i < row; i++ )
    {
      for ( j = 0; j < N_SSR_F_ACTIONS_FORCES; j++ )
	printf( " %g", M[i][j] );
      printf( "\n" );
    }
  printf( "]\n" );
  */

  weight[SSR_F_AHL] = 1e-7;
  weight[SSR_F_AKL] = 1e-7;
  weight[SSR_F_AHR] = 1e-7;
  weight[SSR_F_AKR] = 1e-7;
  weight[SSR_F_AAR] = 1e-7;
  weight[SSR_F_FGSRX] = 1e-7;
  weight[SSR_F_FGSRY] = 1e-7;

  for ( count = 0; count < N_ID_ITERATIONS; count++ )
    {

      for ( i = 0; i < N_ACC; i++ )
	{
	  error[i] = 0;
	  for ( j = 0; j < (N_SSR_F); j++ )
	    error[i] += M[i][j]*forces[j];
	  error[i] += A_qdd_b[i];
	}

      /*
      printf( "ID error: %d:", count );
      for ( j = 0; j < N_ACC; j++ )
	printf( " %g", error[j] );
      printf( "\n" );
      */

      error2 = 0.0;
      for ( j = 0; j < N_ACC; j++ )
	error2 += error[j]*error[j];
      if ( error2 < ID_DS_E2_THRESHOLD )
	break;

      for ( i = 0; i < (N_SSR_F); i++ )
	{
	  gradient[i] = 0;
	  for ( j = 0; j < N_ACC; j++ )
	    gradient[i] += M[j][i]*error[j];
	  update[i] = -epsilon*(gradient[i] + (1e-3)*weight[i]*forces[i]);
	}

      for ( i = 0; i < (N_SSR_F); i++ )
	{
	  forces[i] += update[i];
	}

      df2 = 0.0;
      for ( j = 0; j < N_SSR_F; j++ )
	df2 += (forces[j] - last_forces[j])*(forces[j] - last_forces[j]);
      if ( df2 < ID_DS_F2_THRESHOLD )
	break;
      for ( i = 0; i < N_SSR_F; i++ )
	last_forces[i] = forces[i];
    }

  /*
  printf( "final forces: %d:", count );
  for ( j = 0; j < N_SSR_F; j++ )
    printf( " %g", forces[j] );
  printf( "\n" );
  */

  forces_d[DS_F_AHL] = forces[SSR_F_AHL];
  forces_d[DS_F_AKL] = forces[SSR_F_AKL];
  forces_d[DS_F_AAL] = 0.0;
  forces_d[DS_F_AHR] = forces[SSR_F_AHR];
  forces_d[DS_F_AKR] = forces[SSR_F_AKR];
  forces_d[DS_F_AAR] = forces[SSR_F_AAR];
  forces_d[DS_F_FGSLX] = 0.0;
  forces_d[DS_F_FGSLY] = 0.0;
  forces_d[DS_F_FGSRX] = forces[SSR_F_FGSRX];
  forces_d[DS_F_FGSRY] = forces[SSR_F_FGSRY];
}

/**********************************************************************/
// Approximate version using gradient descent
// assumes desired state is set in d.
// INPUT:
// qdd_d (length N_ACC) has the desired accelerations.
// forces_d (length N_DS_F) has the best guess as to the right forces
//          N_U actions followed by N_C constraint forces.
//          It can be zero.
// OUTPUT:
// forces_d (length N_DS_F) has the desired forces (DS version)

void inverse_dynamics_air( Dynamics *d, double *qdd_d, double *forces_d )
{
  int i, j, k;
  double action[N_U];
  double forces[N_AIR_F];
  double A_qdd[N_ACC]; // A*qdd_d
  double A_qdd_b[N_ACC]; // A*qdd_d - b
  double M[N_ACC][N_AIR_F];
  double error[N_ACC];
  double weight[N_AIR_F];
  double gradient[N_AIR_F];
  double epsilon = 1e-1;
  double update[N_AIR_F];
  double scm, sl, ms, Is, thcm, thl, mth, Ith, tocm, tol, mto, Ito;
  double xh, yh, akl, ahl, akr, ahr, ato;
  int row;
  double last_forces[N_AIR_F];
  double error2;
  double df2;
  int count; 

  // use zero actions to form A, b
  for ( i = 0; i < N_U; i++ )
    action[i] = 0;

  form_matrices_air( d, action ); // form A, b

  // create initial guess as to forces.
  last_forces[AIR_F_AHL] = forces[AIR_F_AHL] = forces_d[DS_F_AHL];
  last_forces[AIR_F_AKL] = forces[AIR_F_AKL] = forces_d[DS_F_AKL];
  last_forces[AIR_F_AHR] = forces[AIR_F_AHR] = forces_d[DS_F_AHR];
  last_forces[AIR_F_AKR] = forces[AIR_F_AKR] = forces_d[DS_F_AKR];

  for ( i = 0; i < N_ACC; i++ )
    {
      A_qdd[i] = 0;
      for ( j = 0; j < N_ACC; j++ )
	A_qdd[i] += d->nr_a[i+1][j+1]*qdd_d[j];
      A_qdd_b[i] =  A_qdd[i] - d->nr_b[i+1];
    }

  /*
  printf( "A_qdd_b: " );
  for ( i = 0; i < N_ACC; i++ )
    printf( "%g ", A_qdd_b[i] );
  printf( "\n" );
  */

  // Now figure out M
  // use Mathematica variable names
  scm = d->shank_cm;
  sl = d->shank_length;
  ms = d->shank_mass;
  Is = d->shank_I;
  thcm = d->thigh_cm;
  thl = d->thigh_length;
  mth = d->thigh_mass;
  Ith = d->thigh_I;
  tocm = d->torso_cm;
  tol = d->torso_length;
  mto = d->torso_mass;
  Ito = d->torso_I;

  // use Mathematica variable names
  xh = d->full_state[X_XH];
  yh = d->full_state[X_YH];
  ato = d->full_state[X_ATO];
  ahl = d->full_state[X_AHL];
  ahr = d->full_state[X_AHR];
  akl = d->full_state[X_AKL];
  akr = d->full_state[X_AKR];

  // Numerical Recipes format: indices start at 1
  for ( i = 0; i < N_ACC; i++ )
    {
      for ( j = 0; j < N_AIR_F; j++ )
	M[i][j] = 0.0;
    }

  row = 0; // Numerical Recipes format: indices start at 1

  // 1 XHDD
  row++;

  // 2 YHDD
  row++;

  // 3 ATODD
  row++;

  // 4 AHLDD
  M[row][AIR_F_AHL] = -1.0;
  row++;

  // 5 AHRDD
  M[row][AIR_F_AHR] = -1.0;
  row++;

  // 6 AKLDD
  M[row][AIR_F_AKL] = 1.0;
  row++;

  // 7 AKRDD
  M[row][AIR_F_AKR] = 1.0;
  row++;

  if ( row != N_ACC )
    {
      fprintf( stderr, "inverse_dynamics_ssr: Row error: %d != %d\n",
	       row, N_ACC );
      exit( -1 );
    }

  /*
  printf( "M = [ " );
  for ( i = 0; i < row; i++ )
    {
      for ( j = 0; j < N_AIR_F_ACTIONS_FORCES; j++ )
	printf( " %g", M[i][j] );
      printf( "\n" );
    }
  printf( "]\n" );
  */

  weight[AIR_F_AHL] = 1e-7;
  weight[AIR_F_AKL] = 1e-7;
  weight[AIR_F_AHR] = 1e-7;
  weight[AIR_F_AKR] = 1e-7;

  for ( count = 0; count < N_ID_ITERATIONS; count++ )
    {

      for ( i = 0; i < N_ACC; i++ )
	{
	  error[i] = 0;
	  for ( j = 0; j < (N_AIR_F); j++ )
	    error[i] += M[i][j]*forces[j];
	  error[i] += A_qdd_b[i];
	}

      /*
      printf( "ID error: %d:", count );
      for ( j = 0; j < N_ACC; j++ )
	printf( " %g", error[j] );
      printf( "\n" );
      */

      error2 = 0.0;
      for ( j = 0; j < N_ACC; j++ )
	error2 += error[j]*error[j];
      if ( error2 < ID_DS_E2_THRESHOLD )
	break;

      for ( i = 0; i < (N_AIR_F); i++ )
	{
	  gradient[i] = 0;
	  for ( j = 0; j < N_ACC; j++ )
	    gradient[i] += M[j][i]*error[j];
	  update[i] = -epsilon*(gradient[i] + (1e-3)*weight[i]*forces[i]);
	}

      for ( i = 0; i < (N_AIR_F); i++ )
	{
	  forces[i] += update[i];
	}

      df2 = 0.0;
      for ( j = 0; j < N_AIR_F; j++ )
	df2 += (forces[j] - last_forces[j])*(forces[j] - last_forces[j]);
      if ( df2 < ID_DS_F2_THRESHOLD )
	break;
      for ( i = 0; i < N_AIR_F; i++ )
	last_forces[i] = forces[i];
    }

  /*
  printf( "final forces: %d:", count );
  for ( j = 0; j < N_AIR_F; j++ )
    printf( " %g", forces[j] );
  printf( "\n" );
  */

  forces_d[DS_F_AHL] = forces[AIR_F_AHL];
  forces_d[DS_F_AKL] = forces[AIR_F_AKL];
  forces_d[DS_F_AAL] = 0.0;
  forces_d[DS_F_AHR] = forces[AIR_F_AHR];
  forces_d[DS_F_AKR] = forces[AIR_F_AKR];
  forces_d[DS_F_AAR] = 0.0;
  forces_d[DS_F_FGSLX] = 0.0;
  forces_d[DS_F_FGSLY] = 0.0;
  forces_d[DS_F_FGSRX] = 0.0;
  forces_d[DS_F_FGSRY] = 0.0;
}

/**********************************************************************/

void inverse_dynamics( Dynamics *d, double *qdd_d, double *forces_d )
{

  switch ( d->dynamics_state )
    {
    case D_DS:
      inverse_dynamics_ds( d, qdd_d, forces_d );
      break;
    case D_SSL:
      inverse_dynamics_ssl( d, qdd_d, forces_d );
      break;
    case D_SSR:
      inverse_dynamics_ssr( d, qdd_d, forces_d );
      break;
    case D_AIR:
      inverse_dynamics_air( d, qdd_d, forces_d );
      break;
    default:
      fprintf( stderr, "Unrecognized dynamics state %d in inverse_dynamics.\n",
	       d->dynamics_state );
      exit( -1 );
      break;
    }
}

/**********************************************************************/
// given configuration, find necessary static forces

statics_ds( Dynamics *d )
{
  int i, j, k;
  double scm, sl, ms, Is, thcm, thl, mth, Ith, tocm, tol, mto, Ito;
  double xh, yh, akl, ahl, akr, ahr, ato;
  int row;
  
  // use Mathematica variable names
  scm = d->shank_cm;
  sl = d->shank_length;
  ms = d->shank_mass;
  Is = d->shank_I;
  thcm = d->thigh_cm;
  thl = d->thigh_length;
  mth = d->thigh_mass;
  Ith = d->thigh_I;
  tocm = d->torso_cm;
  tol = d->torso_length;
  mto = d->torso_mass;
  Ito = d->torso_I;

  // use Mathematica variable names
  xh = d->full_state[X_XH];
  yh = d->full_state[X_YH];
  ato = d->full_state[X_ATO];
  ahl = d->full_state[X_AHL];
  ahr = d->full_state[X_AHR];
  akl = d->full_state[X_AKL];
  akr = d->full_state[X_AKR];

  // Numerical Recipes format: indices start at 1
  for ( i = 1; i <= N_A_B; i++ )
    {
      d->nr_b[i] = 0.0;
      for ( j = 1; j <= N_A_B; j++ )
	d->nr_a[i][j] = 0.0;
    }

  // Solving Ax = b

  row = 1; // Numerical Recipes format: indices start at 1

  // 1 XHDD
  d->nr_a[row][DS_F_FGSLX+1] = -1.0;
  d->nr_a[row][DS_F_FGSRX+1] = -1.0;

  d->nr_b[row] = 
    -(
      0.0
      );

  row++;

  // 2 YHDD
  d->nr_a[row][DS_F_FGSLY+1] = -1.0;
  d->nr_a[row][DS_F_FGSRY+1] = -1.0;

  d->nr_b[row] = 
    -(
      G*(2*(ms + mth) + mto)
      );

  row++;

  // 3 ATODD
  d->nr_a[row][DS_F_AAL+1] = 1.0;
  d->nr_a[row][DS_F_AAR+1] = 1.0;

  d->nr_a[row][DS_F_FGSLX+1] =
    -(thl*Cos(ahl + ato) + sl*Cos(ahl - akl + ato));
  d->nr_a[row][DS_F_FGSLY+1] =
    -(thl*Sin(ahl + ato) + sl*Sin(ahl - akl + ato));
  d->nr_a[row][DS_F_FGSRX+1] =
    -(thl*Cos(ahr + ato) + sl*Cos(ahr - akr + ato));
  d->nr_a[row][DS_F_FGSRY+1] =
    -(thl*Sin(ahr + ato) + sl*Sin(ahr - akr + ato));

  d->nr_b[row] = 
    -(
     G*(-(mto*tocm*Sin(ato)) + (-(mth*thcm) + (ms + mth)*thl)*Sin(ahl + ato) + 
       (-(mth*thcm) + (ms + mth)*thl)*Sin(ahr + ato) - 
       ms*(scm - sl)*(Sin(ahl - akl + ato) + Sin(ahr - akr + ato)))
      );

  row++;

  // 4 AHLDD
  d->nr_a[row][DS_F_AAL+1] = 1.0;
  d->nr_a[row][DS_F_AHL+1] = -1.0;

  d->nr_a[row][DS_F_FGSLX+1] =
    -(thl*Cos(ahl + ato) + sl*Cos(ahl - akl + ato));
  d->nr_a[row][DS_F_FGSLY+1] =
    -(thl*Sin(ahl + ato) + sl*Sin(ahl - akl + ato));

  d->nr_b[row] = 
    -(
      G*(-(mth*thcm) + (ms + mth)*thl)*Sin(ahl + ato) + 
      G*ms*(-scm + sl)*Sin(ahl - akl + ato)
      );

  row++;

  // 5 AHRDD
  d->nr_a[row][DS_F_AAR+1] = 1.0;
  d->nr_a[row][DS_F_AHR+1] = -1.0;

  d->nr_a[row][DS_F_FGSRX+1] =
    -(thl*Cos(ahr + ato) + sl*Cos(ahr - akr + ato));
  d->nr_a[row][DS_F_FGSRY+1] =
    -(thl*Sin(ahr + ato) + sl*Sin(ahr - akr + ato));

  d->nr_b[row] = 
    -(
      G*(-(mth*thcm) + (ms + mth)*thl)*Sin(ahr + ato) + 
      G*ms*(-scm + sl)*Sin(ahr - akr + ato)
      );
  
  row++;

  // 6 AKLDD
  d->nr_a[row][DS_F_AAL+1] = -1.0;
  d->nr_a[row][DS_F_AKL+1] = 1.0;

  d->nr_a[row][DS_F_FGSLX+1] = -(-sl*Cos(ahl - akl + ato));
  d->nr_a[row][DS_F_FGSLY+1] = -(-sl*Sin(ahl - akl + ato));

  d->nr_b[row] = 
    -(
      ms*(scm - sl)*(G*Sin(ahl - akl + ato))
      );

  row++;

  // 7 AKRDD
  d->nr_a[row][DS_F_AAR+1] = -1.0;
  d->nr_a[row][DS_F_AKR+1] = 1.0;

  d->nr_a[row][DS_F_FGSRX+1] = -(-sl*Cos(ahr - akr + ato));
  d->nr_a[row][DS_F_FGSRY+1] = -(-sl*Sin(ahr - akr + ato));

  d->nr_b[row] = 
    -(
      ms*(scm - sl)*(G*Sin(ahr - akr + ato))
      );

  row++;

  if ( row != N_ACC + 1 )
    {
      fprintf( stderr, "Row error: %d != %d\n", row, N_ACC + 1 );
      exit( -1 );
    }

  printf( "Aeq = [ " );
  for ( i = 1; i < row; i++ )
    {
      for ( j = 1; j <= N_DS_F; j++ )
	printf( " %g", d->nr_a[i][j] );
      printf( "\n" );
    }
  printf( "]\n" );

  printf( "beq = [ " );
  for ( i = 1; i < row; i++ )
    printf( "%g\n", d->nr_b[i] );
  printf( "]\n" );

  printf( "Now do quadprog in matlab\n" );

  exit( -1 );
}

/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
  /*
  // Euler method for accelerations and velocities
  accelerations( d, action, xd );

  for ( i = 0; i < N_X; i++ )
    d->next_state[i] = start_state[i] + h*xd[i];

  d->time += d->time_step;
  // End EULER
  */

  /*
  // Euler method for accelerations, modified Euler for velocities
  accelerations( d, action, xd );

  for ( i = 0; i < N_X_ACC; i++ )
    {
      new_velocity = d->state[i + N_X_ACC] + xd[i + N_X_ACC]*d->time_step;
      d->next_state[i] = d->state[i] + d->time_step*(new_velocity + d->state[i + N_X_ACC])/2;
      d->next_state[i + N_X_ACC] = new_velocity;
    }

  d->time += d->time_step;
  // End EULER2
  */
/**********************************************************************/
/**********************************************************************/
// 4th order Runge Kutta method for accelerations and velocities

void runge_kutta4( Dynamics *d, double time_step )
{
  int i, j;
  double start_state[N_X];
  double xd[N_X];
  double k1[N_X], k2[N_X], k3[N_X], k4[N_X];
  double tmp[N_X], state[N_X];

  dv_copy( d->full_state, start_state, N_X );

  // Get k1
  accelerations( d, d->action, xd );
  dv_scale( xd, time_step, k1, N_X );

  // Get k2
  d->time += 0.5*time_step;
  dv_scale( k1, 0.5, tmp, N_X );
  dv_add( start_state, tmp, state, N_X );

  // use new full state
  dv_copy( state, d->full_state, N_X );

  accelerations( d, d->action, xd );

  dv_scale( xd, time_step, k2, N_X );

  // Get k3
  dv_scale( k2, 0.5, tmp, N_X );
  dv_add( start_state, tmp, state, N_X );

  // use new full state
  dv_copy( state, d->full_state, N_X );

  accelerations( d, d->action, xd );

  dv_scale( xd, time_step, k3, N_X );

  // Get k4
  d->time += 0.5*time_step;
  dv_add( start_state, k3, state, N_X );

  // use new full state
  dv_copy( state, d->full_state, N_X );

  accelerations( d, d->action, xd );

  dv_scale( xd, time_step, k4, N_X );

  dv_add( k2, k3, tmp, N_X );
  dv_scale( tmp, 2.0, tmp, N_X );
  dv_acc( k1, tmp, N_X );
  dv_acc( k4, tmp, N_X );
  dv_scale( tmp, 1.0/6.0, tmp, N_X );
  dv_add( start_state, tmp, state, N_X );

  dv_copy( state, d->full_state, N_X );
}

/**********************************************************************/
/**********************************************************************/

void air_to_ds( Dynamics *d )
{
  int i, j, k;
  double action[N_U];
  double am1, am2;

  printf( "air_to_ds %g\n", d->time );

  dv_zero( action, N_U );

  // could just form A matrix
  form_matrices_ds( d, action );

  // form b vector
  for ( i = 1; i <= N_DS_B; i++ )
    d->nr_b[i] = 0;

  for ( i = 0; i < N_ACC; i++ )
    {
      d->nr_b[i+1] = 0;
      for ( k = 1, j = N_ACC; k <= N_ACC; k++, j++ )
	{
	  d->nr_b[i+1] += d->nr_a[i+1][k]*d->full_state[j];
	}
    }

  solve_Ax_b( d );

  printf( "Impulse: %g %g %g %g\n",
	  d->nr_b[DS_B_FGSLX+1], d->nr_b[DS_B_FGSLY+1],
	  d->nr_b[DS_B_FGSRX+1], d->nr_b[DS_B_FGSRY+1] );

  for ( i = 0, j = N_ACC; i < N_ACC; i++, j++ )
    d->full_state[j] = d->nr_b[i+1];

  d->dynamics_state = D_DS;

  forward_kinematics( d );
  /* Only for debugging
  d->lne_energy = compute_energy_lne( d );
  check_energies_lne( d );
  */
  printf( "xyald: %g %g\n", d->xald, d->yald );
  printf( "xyard: %g %g\n", d->xard, d->yard );
}

/**********************************************************************/

void air_to_ssl( Dynamics *d )
{
  int i, j, k;
  double action[N_U];
  double am1, am2;

  printf( "air_to_ssl %g\n", d->time );

  dv_zero( action, N_U );

  // could just form A matrix
  form_matrices_ssl( d, action );

  // form b vector
  for ( i = 1; i <= N_SSL_B; i++ )
    d->nr_b[i] = 0;

  for ( i = 0; i < N_ACC; i++ )
    {
      d->nr_b[i+1] = 0;
      for ( k = 1, j = N_ACC; k <= N_ACC; k++, j++ )
	{
	  d->nr_b[i+1] += d->nr_a[i+1][k]*d->full_state[j];
	}
    }

  /*
  for ( i = 1; i <= N_SSL_B; i++ )
    {
      printf( "a%d: \n", i, d->nr_a[i][j] );
      for ( j = 1; j <= N_SSL_B; j++ )
	printf( " %g", d->nr_a[i][j] );
      printf( "\n" );
    }

  printf( "b: \n" );
  for ( i = 0; i < N_SSL_B; i++ )
    printf( "%g ", d->nr_b[i+1] );
  printf( "\n" );
  */

  solve_Ax_b( d );

  printf( "Impulse: %g %g\n", d->nr_b[DS_B_FGSLX+1], d->nr_b[DS_B_FGSLY+1] );

  /*
  printf( "b after: \n" );
  for ( i = 0; i < N_SSL_B; i++ )
    printf( "%g ", d->nr_b[i+1] );
  printf( "\n" );

  printf( "velocity before: " );
  for ( i = 0, j = N_ACC; i < N_ACC; i++, j++ )
    printf( "%g ", d->full_state[j] );
  printf( "\n" );
  */
  d->am_point[0] = d->xal;
  d->am_point[1] = d->yal;
  am1 = angular_momentum( d );

  for ( i = 0, j = N_ACC; i < N_ACC; i++, j++ )
    d->full_state[j] = d->nr_b[i+1];

  /*
  printf( "velocity after: " );
  for ( i = 0, j = N_ACC; i < N_ACC; i++, j++ )
    printf( "%g ", d->full_state[j] );
  printf( "\n" );
  */

  d->dynamics_state = D_SSL;

  forward_kinematics( d );
  /* Only for debugging
  d->lne_energy = compute_energy_lne( d );
  check_energies_lne( d );
  */
  am2 = angular_momentum( d );
  printf( "angular momentum: %g %g %g\n", am1, am2, am1-am2 );
  printf( "xyald: %g %g\n", d->xald, d->yald );
}

/**********************************************************************/

void air_to_ssr( Dynamics *d )
{
  int i, j, k;
  double action[N_U];
  double am1, am2;

  printf( "air_to_ssr %g\n", d->time );

  dv_zero( action, N_U );

  // could just form A matrix
  form_matrices_ssr( d, action );

  // form b vector
  for ( i = 1; i <= N_SSR_B; i++ )
    d->nr_b[i] = 0;

  for ( i = 0; i < N_ACC; i++ )
    {
      d->nr_b[i+1] = 0;
      for ( k = 1, j = N_ACC; k <= N_ACC; k++, j++ )
	{
	  d->nr_b[i+1] += d->nr_a[i+1][k]*d->full_state[j];
	}
    }

  solve_Ax_b( d );

  printf( "Impulse: %g %g\n", d->nr_b[DS_B_FGSRX+1], d->nr_b[DS_B_FGSRY+1] );

  d->am_point[0] = d->xar;
  d->am_point[1] = d->yar;
  am1 = angular_momentum( d );

  for ( i = 0, j = N_ACC; i < N_ACC; i++, j++ )
    d->full_state[j] = d->nr_b[i+1];

  d->dynamics_state = D_SSR;

  forward_kinematics( d );
  /* Only for debugging
  d->lne_energy = compute_energy_lne( d );
  check_energies_lne( d );
  */
  am2 = angular_momentum( d );
  printf( "angular momentum: %g %g %g\n", am1, am2, am1-am2 );
  printf( "xyard: %g %g\n", d->xard, d->yard );
}

/**********************************************************************/

void ssl_to_ssr( Dynamics *d )
{
  int i, j, k;
  double action[N_U];
  double am1, am2;

  printf( "ssl_to_ssr %g\n", d->time );

  dv_zero( action, N_U );

  // could just form A matrix
  form_matrices_ssr( d, action );

  // form b vector
  for ( i = 1; i <= N_SSR_B; i++ )
    d->nr_b[i] = 0;

  for ( i = 0; i < N_ACC; i++ )
    {
      d->nr_b[i+1] = 0;
      for ( k = 1, j = N_ACC; k <= N_ACC; k++, j++ )
	{
	  d->nr_b[i+1] += d->nr_a[i+1][k]*d->full_state[j];
	}
    }

  solve_Ax_b( d );

  printf( "Impulse: %g %g\n", d->nr_b[DS_B_FGSRX+1], d->nr_b[DS_B_FGSRY+1] );

  d->am_point[0] = d->xar;
  d->am_point[1] = d->yar;
  am1 = angular_momentum( d );

  for ( i = 0, j = N_ACC; i < N_ACC; i++, j++ )
    d->full_state[j] = d->nr_b[i+1];

  d->dynamics_state = D_SSR;

  forward_kinematics( d );
  /* Only for debugging
  d->lne_energy = compute_energy_lne( d );
  check_energies_lne( d );
  */
  am2 = angular_momentum( d );
  printf( "angular momentum: %g %g %g\n", am1, am2, am1-am2 );
  printf( "xyard: %g %g\n", d->xard, d->yard );
}

/**********************************************************************/

void ssl_to_ds( Dynamics *d )
{
  int i, j, k;
  double action[N_U];
  double am1, am2;
  double stash_forces[N_A_B+1];

  printf( "ssl_to_ds %g\n", d->time );

  for ( i = 1; i <= N_A_B; i++ )
    stash_forces[i] = d->nr_b[i];

  printf( "Forces: %g %g\n", stash_forces[SSL_B_FGSLX+1],
	  stash_forces[SSL_B_FGSLY+1] );

  dv_zero( action, N_U );

  // could just form A matrix
  form_matrices_ds( d, action );

  // form b vector
  for ( i = 1; i <= N_DS_B; i++ )
    d->nr_b[i] = 0;

  for ( i = 0; i < N_ACC; i++ )
    {
      d->nr_b[i+1] = 0;
      for ( k = 1, j = N_ACC; k <= N_ACC; k++, j++ )
	{
	  d->nr_b[i+1] += d->nr_a[i+1][k]*d->full_state[j];
	}
    }

  solve_Ax_b( d );

  printf( "Impulse: %g %g %g %g\n",
	  d->nr_b[DS_B_FGSLX+1], d->nr_b[DS_B_FGSLY+1],
	  d->nr_b[DS_B_FGSRX+1], d->nr_b[DS_B_FGSRY+1] );

  /*
  printf( "velocity before: " );
  for ( i = 0, j = N_ACC; i < N_ACC; i++, j++ )
    printf( "%g ", d->full_state[j] );
  printf( "\n" );
  */
  d->am_point[0] = d->xar;
  d->am_point[1] = d->yar;
  am1 = angular_momentum( d );

  for ( i = 0, j = N_ACC; i < N_ACC; i++, j++ )
    d->full_state[j] = d->nr_b[i+1];
  // combine impulse and forces
  for ( i = DS_B_FGSLX+1; i <= DS_B_FGSRY+1; i++ )
    d->nr_b[i] = d->nr_b[i]/d->time_step;
  d->nr_b[DS_B_FGSLX+1] += stash_forces[SSL_B_FGSLX+1];
  d->nr_b[DS_B_FGSLY+1] += stash_forces[SSL_B_FGSLY+1];

  printf( "New forces: %g %g %g %g\n",
	  d->nr_b[DS_B_FGSLX+1], d->nr_b[DS_B_FGSLY+1],
	  d->nr_b[DS_B_FGSRX+1], d->nr_b[DS_B_FGSRY+1] );

  /*
  printf( "velocity after: " );
  for ( i = 0, j = N_ACC; i < N_ACC; i++, j++ )
    printf( "%g ", d->full_state[j] );
  printf( "\n" );
  */

  d->dynamics_state = D_DS;

  forward_kinematics( d );
  /* Only for debugging
  d->lne_energy = compute_energy_lne( d );
  check_energies_lne( d );
  */
  am2 = angular_momentum( d );
  printf( "angular momentum: %g %g %g: %g\n", am1, am2,
	  (d->xal - d->xar)*d->nr_b[DS_B_FGSLY+1]
	  - (d->yal - d->yar)*d->nr_b[DS_B_FGSLX+1],
	  am1 - am2 + ((d->xal - d->xar)*d->nr_b[DS_B_FGSLY+1]
		       - (d->yal - d->yar)*d->nr_b[DS_B_FGSLX+1]) );
  printf( "xyald: %g %g\n", d->xald, d->yald );
  printf( "xyard: %g %g\n", d->xard, d->yard );
}

/**********************************************************************/

void ssr_to_ssl( Dynamics *d )
{
  int i, j, k;
  double action[N_U];
  double am1, am2;

  printf( "ssr_to_ssl %g\n", d->time );

  dv_zero( action, N_U );

  // could just form A matrix
  form_matrices_ssl( d, action );

  // form b vector
  for ( i = 1; i <= N_SSL_B; i++ )
    d->nr_b[i] = 0;

  for ( i = 0; i < N_ACC; i++ )
    {
      d->nr_b[i+1] = 0;
      for ( k = 1, j = N_ACC; k <= N_ACC; k++, j++ )
	{
	  d->nr_b[i+1] += d->nr_a[i+1][k]*d->full_state[j];
	}
    }

  solve_Ax_b( d );

  printf( "Impulse: %g %g\n", d->nr_b[DS_B_FGSLX+1], d->nr_b[DS_B_FGSLY+1] );

  d->am_point[0] = d->xal;
  d->am_point[1] = d->yal;
  am1 = angular_momentum( d );

  for ( i = 0, j = N_ACC; i < N_ACC; i++, j++ )
    d->full_state[j] = d->nr_b[i+1];

  d->dynamics_state = D_SSL;

  forward_kinematics( d );
  /* Only for debugging
  d->lne_energy = compute_energy_lne( d );
  check_energies_lne( d );
  */
  am2 = angular_momentum( d );
  printf( "angular momentum: %g %g %g\n", am1, am2, am1-am2 );
  printf( "xyald: %g %g\n", d->xald, d->yald );
}

/**********************************************************************/

void ssr_to_ds( Dynamics *d )
{
  int i, j, k;
  double action[N_U];
  double am1, am2;
  double stash_forces[N_A_B+1];

  printf( "ssr_to_ds %g\n", d->time );

  for ( i = 1; i <= N_A_B; i++ )
    stash_forces[i] = d->nr_b[i];

  printf( "Forces: %g %g\n", stash_forces[SSR_B_FGSRX+1],
	  stash_forces[SSR_B_FGSRY+1] );

  dv_zero( action, N_U );

  // could just form A matrix
  form_matrices_ds( d, action );

  // form b vector
  for ( i = 1; i <= N_DS_B; i++ )
    d->nr_b[i] = 0;

  for ( i = 0; i < N_ACC; i++ )
    {
      d->nr_b[i+1] = 0;
      for ( k = 1, j = N_ACC; k <= N_ACC; k++, j++ )
	{
	  d->nr_b[i+1] += d->nr_a[i+1][k]*d->full_state[j];
	}
    }

  solve_Ax_b( d );

  printf( "Impulse: %g %g %g %g\n",
	  d->nr_b[DS_B_FGSLX+1], d->nr_b[DS_B_FGSLY+1],
	  d->nr_b[DS_B_FGSRX+1], d->nr_b[DS_B_FGSRY+1] );

  d->am_point[0] = d->xal;
  d->am_point[1] = d->yal;
  am1 = angular_momentum( d );

  for ( i = 0, j = N_ACC; i < N_ACC; i++, j++ )
    d->full_state[j] = d->nr_b[i+1];
  // combine impulse and forces
  for ( i = DS_B_FGSLX+1; i <= DS_B_FGSRY+1; i++ )
    d->nr_b[i] = d->nr_b[i]/d->time_step;
  d->nr_b[DS_B_FGSRX+1] += stash_forces[SSR_B_FGSRX+1];
  d->nr_b[DS_B_FGSRY+1] += stash_forces[SSR_B_FGSRY+1];

  printf( "New forces: %g %g %g %g\n",
	  d->nr_b[DS_B_FGSLX+1], d->nr_b[DS_B_FGSLY+1],
	  d->nr_b[DS_B_FGSRX+1], d->nr_b[DS_B_FGSRY+1] );

  d->dynamics_state = D_DS;

  forward_kinematics( d );
  /* Only for debugging
  d->lne_energy = compute_energy_lne( d );
  check_energies_lne( d );
  */
  am2 = angular_momentum( d );
  printf( "angular momentum: %g %g %g: %g\n", am1, am2,
	  (d->xar - d->xal)*d->nr_b[DS_B_FGSRY+1]
	  - (d->yar - d->yal)*d->nr_b[DS_B_FGSRX+1],
	  am1 - am2 + ((d->xar - d->xal)*d->nr_b[DS_B_FGSRY+1]
		       - (d->yar - d->yal)*d->nr_b[DS_B_FGSRX+1]) );
  printf( "xyald: %g %g\n", d->xald, d->yald );
  printf( "xyard: %g %g\n", d->xard, d->yard );
}

/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
// don't worry about sliding (yet).

void identify_dynamics_state( Dynamics *d )
{
  switch ( d->dynamics_state )
    {
    case D_AIR:
      if ( (d->yal <= 0 && d->yald < 0 )
	   && (d->yar <= 0 && d->yard < 0 ) )
	{
	  air_to_ds( d );
	  break;
	}
      if ( d->yal <= 0 && d->yald < 0 )
	{
	  air_to_ssl( d );
	  break;
	}
      if ( d->yar <= 0 && d->yard < 0 )
	{
	  air_to_ssr( d );
	  break;
	}
      break;
    case D_SSL:
      if ( ( d->s_fy_l <= 0.0 ) && (d->yar <= 0 && d->yard < 0 ) )
	{
	  ssl_to_ssr( d );
	  break;
	}
      if ( ( d->s_fy_l <= 0.0 ) )
	{
	  d->dynamics_state = D_AIR;
	  break;
	}
      if ( (d->yar <= 0 && d->yard < 0 ) )
	{
	  ssl_to_ds( d );
	  break;
	}
      break;
    case D_SSR:
      if ( ( d->s_fy_r <= 0.0 ) && (d->yal <= 0 && d->yald < 0 ) )
	{
	  ssr_to_ssl( d );
	  break;
	}
      if ( ( d->s_fy_r <= 0.0 ) )
	{
	  d->dynamics_state = D_AIR;
	  break;
	}
      if ( (d->yal <= 0 && d->yald < 0 ) )
	{
	  ssr_to_ds( d );
	  break;
	}
      break;
    case D_DS:
      if ( ( d->s_fy_l <= 0.0 ) && ( d->s_fy_r <= 0.0 ) )
	{
	  d->dynamics_state = D_AIR;
	  break;
	}
      if ( ( d->s_fy_l <= 0.0 ) )
	{
	  d->dynamics_state = D_SSR;
	  break;
	}
      if ( ( d->s_fy_r <= 0.0 ) )
	{
	  d->dynamics_state = D_SSL;
	  break;
	}
      break;
    default:
      fprintf( stderr,
	       "Unrecognized dynamics state %d in identify_dynamics_state.\n",
	       d->dynamics_state );
      exit( -1 );
      break;
    }
}

/**********************************************************************/
/**********************************************************************/

int integrate_one_step( Dynamics *d, double *action, double *next_state )
{
  double xd[N_X];

  if ( action != NULL )
    dv_copy( action, d->action, N_U ); // stash action in data structure.

  d->last_dynamics_state = d->dynamics_state;
  dv_copy( d->full_state, d->last_full_state, N_X );

  runge_kutta4( d, d->time_step );

  // use new full state
  forward_kinematics( d );
  /* Only for debugging
  d->lne_energy = compute_energy_lne( d );
  check_energies_lne( d );
  */

  // handle change of dynamics_state
  // currently, don't back up.
  identify_dynamics_state( d );

  // make sure up to date accelerations stored with old actions in data file.
  accelerations( d, d->action, xd ); 

  return 0;
}

/**********************************************************************/
/**********************************************************************/

int integrate_one_step_alt( Dynamics *d,
			    double *state,
			    double *action,
			    int index,
			    double *next_state )
{
  // set_state( d, 0.0, state );
  fprintf( stderr, "integrate_one_step_alt not implemented\n" );
  exit( -1 );
  integrate_one_step( d, action, next_state );
  return 0;
}

/**********************************************************************/

double one_step_cost( Dynamics *d, double *state, double *action, int index )
{
  double ds[N_X], s2[N_X];
  double cx, cu;
  double u2[N_U];

  /*
  dv_subtract( state, d->desired_state, ds, N_X );
  dmv_mult( d->qq, ds, s2, N_X, N_X );
  dv_dot( ds, s2, &cx, N_X );

  dmv_mult( d->rr, action, u2, N_U, N_U );
  dv_dot( action, u2, &cu, N_U );

  return 0.5*(cx + cu);
  */
  fprintf( stderr, "one_step_cost() not implemented.\n" );
  exit( -1 );

  return 1.0;
}

/**********************************************************************/
// return TRUE if crash, FALSE otherwise.

int crash( Dynamics *d )
{
  int i;

#ifdef COMMENT
  // humanoid too low
  if ( d->yh < 0 || d->yhead < 0 || d->ykl < 0 || 0 )
    return CRASH_TOO_LOW;

  // check knees bending backwards
  if ( d->state[X_AKL] < 0.0 || d->state[X_AKR] < 0.0 )
    return CRASH_KNEE_BACKWARDS;

  // ground/left-foot contact force
  if ( d->nr_b_ne[NE_FGSLY+1] < 0.0 )
    return CRASH_GSL_FORCE_BAD;
  // ground/right-foot contact force
  if ( d->nr_b_ne[NE_FGSLY+1] < 0.0 )
    return CRASH_GSR_FORCE_BAD;

  // Friction cones should not be violated (no slipping contacts).
  // board/left-foot
  if ( fabs( d->nr_b_ne[NE_FGSLX+1]/d->nr_b_ne[NE_FGSLY+1] ) > 1.0 )
    return CRASH_GSL_SLIP;
  // board/right-foot
  if ( fabs( d->nr_b_ne[NE_FGSRX+1]/d->nr_b_ne[NE_FGSRY+1] ) > 1.0 )
    return CRASH_GSR_SLIP;

  // The maximum actuator torques can not be exceeded.
  for ( i = 0; i < N_U; i++ )
    if ( fabs( d->action[i] ) > U_MAX )
      return CRASH_TORQUE_TOO_BIG;
#endif

  return NOT_CRASHED; // 0
}

/**********************************************************************/

const char *crash_messages[] =
  {
    "NOT CRASHED",
    "CRASH_TOO_LOW",
    "CRASH_GSL_FORCE_BAD",
    "CRASH_GSR_FORCE_BAD",
    "CRASH_GSL_SLIP",
    "CRASH_GSR_SLIP",
    "CRASH_KNEE_BACKWARDS",
    "CRASH_TORQUE_TOO_BIG"
  };

void print_crash_message( Dynamics *d, int error_number )
{
  if ( d->status == error_number )
    return;

  if ( error_number <= FATAL_CRASH && error_number > 0 )
    printf( "FATAL: %g: %d %s\n", d->time, error_number,
	    crash_messages[ error_number ] );
  else
    printf( "%g: %d %s\n", d->time, error_number,
	    crash_messages[ error_number ] );
}

/**********************************************************************/


